#!/usr/bin/python
# coding: utf-8

###############################################################################
#
# This code was originally written by JP Francoia for the Cronin Group.
# I (JMP) have been given the task of continuing it.
# I plan to make minimal changes, but I have added a lot of comments to help me
# Contact me at juanma@chem.gla.ac.uk
#
###############################################################################

import sys
from PyQt5 import QtGui, QtSql, QtCore, QtWidgets
import logging
import datetime
import os
import time
import shutil
import webbrowser
from typing import Dict, List

# Personal modules
from chemscad.log import MyLog
import chemscad.functions as functions
import chemscad.app_constants as app_constants
from chemscad.stl_viewer import STLViewer
from chemscad.little_thread import LittleThread

# Dialog boxes
from chemscad.choice_module import ChoiceModule
from chemscad.align_tops import AlignTops
from chemscad.align_filters import AlignFilters
from chemscad.settings import Settings
from chemscad.filtersettings import FilterSettings
from chemscad.help_info import help_info

# from composite import Composite
from chemscad.dashboard import Dashboard

# CAD modules
from ccad.object import ObjectCAD
from ccad.connector import ConnectorCAD
from ccad import Assembly
from ccad import ReactorCAD


# # DEBUG: do not show deprecation warningmport warningss
# import warnings
# warnings.filterwarnings("ignore", category=DeprecationWarning)

class MyWindow(QtWidgets.QMainWindow):
    def __init__(self):

        super(MyWindow, self).__init__()

        # Get the DATA_PATH and the resource_dir pathes
        self.resource_dir, self.DATA_PATH = functions.getRightDirs()

        # create a render dir for the user
        self.createUserDirs()

        # Create the logger w/ the appropriate size
        self.l = MyLog("activity.log")

        # Set the logging level
        self.l.setLevel(logging.DEBUG)

        # Log the resource and data directories and the library path
        # fn and fs are chemcad parameters related to the quality of the render
        self.l.info(f"resource_dir: {self.resource_dir}")
        self.l.info(f"DATA_PATH: {self.DATA_PATH}")
        self.l.info(f"fn: {app_constants.fn}")
        self.l.info(f"fs: {app_constants.fs}")
        self.l.info(QtWidgets.QApplication.libraryPaths())
        self.l.info("Starting the program")

        # Clean the renders dir from any file
        # He does this again 10 lines ahead. I dont see why it needs to be repeated twice. Commenting for now
        # functions.cleanRendersDir(self.l)

        # Start a timer
        start_time = datetime.datetime.now()
        diff_time = start_time

        # Clean the renders dir, in case the software crashed before and didn't do it
        functions.cleanRendersDir(self.l)

        # Object to store options and preferences
        self.options = QtCore.QSettings("options.ini", QtCore.QSettings.IniFormat)

        # Prepare stack for undo
        self.undoStack = QtWidgets.QUndoStack()

        # Define in self different QT actions such as close program, render,...
        self.defineActions()

        # Create the GUI
        self.initUI()
        self.l.debug(f"initUI took {datetime.datetime.now() - diff_time}")

        diff_time = datetime.datetime.now()

        # Define the slots
        QtWidgets.qApp.processEvents()
        self.defineSlots()
        self.l.debug(f"defineSlots took {datetime.datetime.now() - diff_time}")
        diff_time = datetime.datetime.now()

        # Restore the settings
        QtWidgets.qApp.processEvents()
        self.restoreSettings()
        self.l.debug(f"restoreSettings took {datetime.datetime.now() - diff_time}")
        diff_time = datetime.datetime.now()

        QtWidgets.qApp.processEvents()

        self.l.info(f"Boot took {datetime.datetime.now() - start_time}")

        # Dict to store available modules
        self.dic_available_modules = self.listModules()

        # Create an assembly object to store the chem_modules
        self.ass = Assembly(renders_dir=self.renders_dir, logger=self.l)


        # Bool to check if assembly is rendered (unmodified + rendered)
        # Can't export STL if False
        self.rendered = False

        self.show()

    def createUserDirs(self):

        """Create the necessary user's dirs"""

        # Rendering dir
        self.renders_dir = os.path.join(self.DATA_PATH, "renders")
        os.makedirs(self.renders_dir, exist_ok=True)

    def defineActions(self):

        """Define software actions"""

        # Action to create new project (clear canvas) #HPD - 22/11/19
        self.newProjectAction = QtWidgets.QAction("&New project", self)
        self.newProjectAction.setShortcut("Ctrl+N")
        self.newProjectAction.setStatusTip("Start new project")
        self.newProjectAction.triggered.connect(self.newProject)

        # Action to quit
        self.exitAction = QtWidgets.QAction("&Quit", self)
        self.exitAction.setShortcut("Ctrl+Q")
        self.exitAction.setStatusTip("Quit")
        self.exitAction.triggered.connect(self.close) #HPD: 19/02/20 - fixed crash when quitting

        # Action to render assembly at high def, for production
        self.renderProdAction = QtWidgets.QAction("&Render for production", self)
        self.renderProdAction.setShortcut("Ctrl+R")
        self.renderProdAction.setStatusTip(
            "Render assembly at high definition, for production"
        )
        self.renderProdAction.triggered.connect(self.renderProd)

        # Action to export assembly as STL
        # self.exportSTLAction = QtWidgets.QAction("&Export as STL", self)
        # self.exportSTLAction.setShortcut("Ctrl+Shift+S")
        # self.exportSTLAction.setStatusTip("Export assembly as STL")
        # self.exportSTLAction.triggered.connect(self.exportSTL)

        # Action to take a screenshot of the stl viewer
        self.screenshotAction = QtWidgets.QAction("Export as image", self)
        self.screenshotAction.setStatusTip("Export an image of the current view")
        self.screenshotAction.triggered.connect(self.screenshot)

        # Action to save current project
        self.saveProjectAction = QtWidgets.QAction("&Save project", self)
        self.saveProjectAction.setShortcut("Ctrl+S")
        self.saveProjectAction.setStatusTip("Save current project")
        self.saveProjectAction.triggered.connect(self.saveProject)

        # Action to import a project (.ccad)
        self.openProjectAction = QtWidgets.QAction("&Open project", self)
        self.openProjectAction.setShortcut("Ctrl+O")
        self.openProjectAction.setStatusTip("Open a CCAD project")
        self.openProjectAction.triggered.connect(self.openProject)

        # Action to set the preferences for assembly angles
        self.preferencesAction = QtWidgets.QAction("&Assembly angle settings", self)
        self.preferencesAction.setStatusTip("Open the ChemSCAD's preferences for assembly angle")
        self.preferencesAction.triggered.connect(lambda: Settings(self))

        # Action to undo
        self.undoAction = self.undoStack.createUndoAction(self, "&Undo")
        self.undoAction.setShortcut("Ctrl+Z")
        self.undoAction.setStatusTip("Undo the last action")


        # Action to redo
        self.redoAction = self.undoStack.createRedoAction(self, "&Redo")
        self.redoAction.setShortcut("Ctrl+Y")
        self.redoAction.setStatusTip("Redo the last action")

        # Action to view the filter settings table for FR and DFR modules
        self.filterSettingsAction = QtWidgets.QAction("&Filter settings", self)
        self.filterSettingsAction.setStatusTip("Open the filter settings table")
        self.filterSettingsAction.triggered.connect(lambda: FilterSettings(self))

        # Action to open Git database
        self.openDatabaseAction = QtWidgets.QAction("&Go to database", self)
        self.openDatabaseAction.setStatusTip("Open the Git database")
        self.openDatabaseAction.setShortcut("Ctrl+D")
        self.openDatabaseAction.triggered.connect(self.openDatabaseURL)

        # Action to view the 'About ChemSCAD' window
        #self.aboutAction = QtWidgets.QAction("&About ChemSCAD", self)
        #self.aboutAction.setStatusTip("See information about ChemSCAD")
        #self.aboutAction.triggered.connect(lambda: help_info(self))

        # Action to open help documentation - link to ChemSCAD Git repository
        self.helpAction = QtWidgets.QAction("&Go to the ChemSCAD documentation", self)
        self.helpAction.setStatusTip("Open the ChemSCAD documentation")
        self.helpAction.setShortcut("Ctrl+H")
        self.helpAction.triggered.connect(self.openDocumentationURL)

        # Action separator
        self.separatorAction = QtWidgets.QAction(self)
        self.separatorAction.setSeparator(True)

    def openDatabaseURL(self): # function to open Git database, can also use lambda fucntion with webbrowser.open('url')
        url = QtCore.QUrl('https://gitlab.com/croningroup/reactionware/reactionware_files/-/tree/master') # link to database goes here
        if not QtGui.QDesktopServices.openUrl(url):
            QtGui.QMessageBox.warning(self, 'Open Url', 'Could not open url')

    def openDocumentationURL(self): # function to open ChemSCAD Git repository (documentation in future), can also use lambda fucntion with webbrowser.open('url')
        url = QtCore.QUrl('https://gitlab.com/croningroup/reactionware/chemscad') # link to ChemSCAD repository (documentation)
        if not QtGui.QDesktopServices.openUrl(url):
            QtGui.QMessageBox.warning(self, 'Open Url', 'Could not open url')

    def closeEvent(self, event):

        """
        Method to perform actions before exiting.
        Allows to save the prefs in a file
        """

        # Clean the renders dir from any file
        functions.cleanRendersDir(self.l)

        # Record the window state and appearance
        self.options.beginGroup("Window")

        # Reinitializing the keys
        self.options.remove("")

        self.l.debug("Saving windows state")
        self.options.setValue("window_geometry", self.saveGeometry())
        self.options.setValue("window_state", self.saveState())

        # # Save the state of the window's splitter
        # self.options.setValue("split_exp", self.split_exp.saveState())

        self.options.endGroup()

        # Be sure ini files finished their tasks
        self.options.sync()

        close = QtWidgets.QMessageBox.question(self,
                                         "QUIT",
                                         "Are you sure want to close this project?",
                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if close == QtWidgets.QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()




    def restoreSettings(self):

        """Restore the prefs of the window"""

        # If windows settings are available, import and use them
        if "Window" in self.options.childGroups():
            self.restoreGeometry(self.options.value("Window/window_geometry"))
            self.restoreState(self.options.value("Window/window_state"))
            # self.split_ana.restoreState(self.options.value("Window/split_ana"))

            # self.split_display_ana.restoreState(
            # self.options.value("Window/split_display_ana"))

    def defineSlots(self):

        """Connect the slots"""

        # self.splitter2.splitterMoved.connect(self.updateCellSize)
        # new_size = self.splitter2.sizes()[0]

        # Add module button connected to dialog box for module choice
        self.button_add_module.clicked.connect(lambda: ChoiceModule(self))
        self.button_add_module.setToolTip('Add a new module')
        # # Button to display main view (preview)
        # self.button_main_view.clicked.connect(lambda: self.displayView("preview.png"))

        # # Display left view (preview)
        # self.button_left_view.clicked.connect(lambda: self.displayView("preview_left.png"))

        # # Display right view (preview)
        # self.button_right_view.clicked.connect(lambda: self.displayView("preview_right.png"))

        # # Display top view (preview)
        # self.button_top_view.clicked.connect(lambda: self.displayView("preview_top.png"))

        # # Display front view (preview)
        # self.button_front_view.clicked.connect(lambda: self.displayView("preview_front.png"))

        # # Display rear view (preview)
        # self.button_rear_view.clicked.connect(lambda: self.displayView("preview_rear.png"))

        # # Display bottom view (preview)
        # self.button_bottom_view.clicked.connect(lambda: self.displayView("preview_bottom.png"))

        # # Render STL files (fs=75 for HD)
        self.button_render.clicked.connect(self.renderProd)
        self.button_render.setToolTip('Render STL')

        # Render Ultra High Res STL files for publication images/figures (fs=200 for UHD)
        self.button_renderUHD.clicked.connect(self.renderProd_UHD)
        self.button_renderUHD.setToolTip('Render UHD STL')

        # Dialog box to select reactors tops to align
        self.button_align_tops.clicked.connect(lambda: AlignTops(self))
        self.button_align_tops.setToolTip('Align reactor tops')

        # Dialog box to select reactors with filters to align
        self.button_align_filters.clicked.connect(lambda: AlignFilters(self))
        self.button_align_filters.setToolTip('Align reactors with filters')

        # Toggle translucent state for mesh in stl viewer
        self.button_translucent.clicked.connect(self.toggleTranslucent)
        self.button_translucent.setToolTip('Toggle X-Ray view')

        # Set the perspective state in STL viewer
        self.button_perspective.clicked.connect(self.togglePerspective)
        self.button_perspective.setToolTip('Toggle perspective view')

        # Save the current project using button
        self.button_save.clicked.connect(self.saveProject)
        self.button_save.setToolTip('Save the current project')

        # Button to toggle the linear assembly state
        self.button_linear.clicked.connect(self.toggleLinearAssembly)
        self.button_linear.setToolTip('Toggle linear assembly')

        # # Undo action
        self.button_undo.clicked.connect(self.undoAction.trigger) #HPD: undo button now working - 19/02/20
        self.button_undo.setToolTip('Undo action')

        # # Redo action
        self.button_redo.clicked.connect(self.redoAction.trigger) #HPD: redo button now working - 19/02/20
        self.button_redo.setToolTip('Redo action')

    def listModules(self) -> Dict[str, str]:

        """
        Build the list of modules by getting names from all name.txt files
        in module_* folders
        """

        dic_modules: Dict[str, str] = dict()

        # In modules/ look for all directories starting with module_
        for entry in os.scandir(os.path.join(self.resource_dir, "gui_modules")):

            if (
                entry.name.startswith("module_") or entry.name.startswith("connector_")
            ) and entry.is_dir():

                # Get the name of the module
                with open(
                    os.path.join(
                        self.resource_dir, "gui_modules", entry.name, "name.txt"
                    ),
                    "r",
                ) as f:
                    name = f.read().strip()

                path_dir_module = os.path.join(
                    self.resource_dir, "gui_modules", entry.name
                )

                # Add the path of the module to PATH
                # Necessary to open a project if each module was initialized
                # by opening the dialog box when creating
                sys.path.append(path_dir_module)

                dic_modules[name] = entry.name

        return dic_modules

    def buildModule(self, cad_object: ObjectCAD):

        """
        ChemModule objects will call this method from their own buildModule
        method. This method handles the display of the new ChemModule, not the
        creation/calculations of the module
        """

        # Now we are using the QtUndo pattern
        command = CommandBuild(cad_object, self.ass, self)
        self.undoStack.push(command)

        # Add the CAD object to assembly
        # Assembly will refresh itself
        # self.ass.appendModule(cad_object)
        # self.refresh()

    def deleteModule(self, cad_object: ObjectCAD):

        """
        ChemModule objects will call this method when deleted.
        This method will delete the module from the assembly.
        """

        # Now we are using the QtUndo pattern
        command = CommandDelete(cad_object, self.ass, self)
        self.undoStack.push(command)
        # # Delete the CAD object from assembly
        # # Assembly will refresh itself
        # self.ass.deleteModule(cad_object)
        # self.refresh()

        # # remove the mesh from the stl_viewer
        # self.stl_viewer.deleteMesh(cad_object.internal_id)
        # self.refresh()

    def openModuleDialog(self, cad_object: ObjectCAD):

        """Open module's dialog box to update parameters"""

        # Import module
        module = __import__(self.dic_available_modules[cad_object.module_type])

        # Build the ChemModule class contained in the module
        # Display the module's window
        module.ChemModule(self, cad_obj=cad_object)

    def refresh(self):

        """
        Called by module_*. Refresh the assembly and the preview when a
        module changes (user entered new parameters)
        """

        # Something changed in the assembly, switching rendered state to False.
        # Can't export stl
        self.rendered = False

        # Refresh assembly
        self.ass.refresh()

        # Refresh dashboard
        self.dashboard.refresh()

        self.renderSTLs()

    def renderSTLs(self):

        """
        Will trigger the rendering of the STLs, for each module.
        These renderings will be started separately in Qt threads
        """

        # TEST
        # return

        self.l.debug("Rendering model")

        # Generathe the scad files (one for each module. generatePreviews
        # generates only one for the entire assembly)
        self.ass.prepareRenderSTLs()

        # Dict to store workers
        # key: file name (necessary to trigger display in STL viewer)
        # value: worker
        self.dict_workers: Dict[str, LittleThread] = dict()

        # Start a timer
        self.render_start_time = datetime.datetime.now()

        # Get the name of the files to be rendered
        list_stl_to_render: List[str] = self.getSTLsToRender()

        # Progress bar to display the progress of the rendering
        self.prog_render = QtWidgets.QProgressDialog(self)
        self.prog_render.setWindowTitle("Rendering modules")
        self.prog_render.setModal(True)
        self.prog_render.setMinimum(0)
        self.prog_render.setMaximum(len(list_stl_to_render))
        self.prog_render.setValue(0)
        self.prog_render.setLabelText(f"0/{len(list_stl_to_render)} modules rendered")
        self.prog_render.show()

        # For each of the scad file, start a thread, and put the thread in the
        # worker dict
        for name in list_stl_to_render:
            worker = LittleThread(self.ass.renderSTL, name)
            self.dict_workers[name] = worker
            worker.start()

        self.checkRenderingStage()

    def renderProd(self):

        """
        Method to render assembly at high definition, for production. Wrappe
        for renderSTLs. Changes assembly.fn before and after rendering - fs = 75
        """

        self.l.debug("Start rendering for production")

        # Tell the assembly to render in HD
        self.ass.switchToHD()

        # Render all modules in HD
        self.renderSTLs()

        # Tell the assembly to render low definition
        self.ass.switchToLD()

        self.l.debug("End endering for production")

    def renderProd_UHD(self, event):

        """
        Method to render assembly at ULTRA high definition, for publication. Wrappe
        for renderSTLs. Changes assembly.fn before and after rendering - fs = 200
        """

        self.l.debug("Start rendering for publicaton")
        rendering_warning = QtWidgets.QMessageBox.question(self,
                                         "WARNING",
                                         "Are you sure want to render the model in UHD? Rendering times can take 10x longer than HD!",
                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if rendering_warning == QtWidgets.QMessageBox.Yes:
            # Tell the assembly to render in UHD
            self.ass.switchToUHD()

            # Render all modules in UHD
            self.renderSTLs()

            # Tell the assembly to render low definition
            self.ass.switchToLD()

            self.l.debug("End rendering for publication")


    def getSTLsToRender(self) -> List[str]:

        """getSTLsToRender

        Returns:
            List[str]: a list of file names to be rendered. Not path, just the name
        """

        self.l.debug("Entering getSTLsToRender")

        list_stl_to_render: List[str] = list()

        for entry in os.scandir(self.renders_dir):

            # Only render "obj_*_to_render.scad" files
            if "_to_render" in entry.name:
                list_stl_to_render.append(entry.name)
                self.l.debug(f"Adding {entry.name} to list of STLs to render")

        return list_stl_to_render

    def checkRenderingStage(self):

        """
        Method called each time a rendering thread finished. Update the
        rendering progress bar, and triggers actions when all threads
        are finished
        """

        # List to store bools. When len(list) = len(dict_workers), all workers
        # are finished. Workers' state can't be checked on the fly, otherwise
        # some instructions might not be executed
        list_bool_finished: List[bool] = list()

        # Freeze the nbr of modules
        nbr_workers = len(self.dict_workers)

        while len(list_bool_finished) != nbr_workers:

            # Update app while workers running
            QtWidgets.qApp.processEvents()

            # Check if workers are still running
            for file_name, worker in list(self.dict_workers.items()):

                if worker.isFinished():

                    list_bool_finished.append(True)
                    del self.dict_workers[file_name]

                    # Load the newly rendered STL file in the STL viewer
                    self.viewSTL(file_name.replace("_to_render", ""))

                    # Update the progress bar
                    nbr_stls_rendered = self.prog_render.value() + 1
                    self.prog_render.setValue(nbr_stls_rendered)
                    self.prog_render.setLabelText(
                        f"{nbr_stls_rendered}/{nbr_workers} modules rendered"
                    )

            time.sleep(0.2)

        total_time = datetime.datetime.now() - self.render_start_time
        self.l.debug(f"All renderings & displays finished in {total_time}")

        # Switching the rendering state to True. Can now export stl
        self.rendered = True

        # Delete the progress bar
        self.prog_render.reset()

        del self.dict_workers
        del self.render_start_time

        # Refresh the STL check boxes (delete and rebuild them) in the STL viewer
        # control center
        self.refreshSTLCheckBoxes()

    def refreshSTLCheckBoxes(self):

        """
        Create a checkbox for each module in assembly, connect it to a method to handle
        state toggling, and add it to the control center (in a dedicated scroll area)
        """

        # Clear all the checkboxes
        self.clearLayout(self.vbox_control_center)

        # Get all the modules in the assembly
        modules_names = list(self.ass.reactors_names.keys()) + list(
            self.ass.connectors_names.keys()
        )

        # Create a checkbox for each module
        for module_name in modules_names:
            checkbox = QtWidgets.QCheckBox(module_name, self.scroll_control_center)
            checkbox.setChecked(True)
            checkbox.stateChanged.connect(self.checkBoxSTLToggled)
            self.vbox_control_center.addWidget(checkbox)

        # Necessary here, otherwise checkboxes don't appear
        self.scroll_control_center.setWidget(self.scrolling_control_center)

    def checkBoxSTLToggled(self):

        """
        A checkbox for each module is present in the control center.
        If the checkbox is checked, the module is displayed in the STL viewer.
        If not, the module is not displayed.

        In the future, a module not displayed will not be exported when user exports
        the final STL
        """

        # Get the module short name from the checkbox that triggered the signal
        module_id = self.ass.getModuleIDFromName(self.sender().text())
        self.stl_viewer.toggleVisibility(module_id)

    def screenshot(self):

        """Ask the STL viewere to take a screenshot"""

        # Let user choose destination file
        dst = QtWidgets.QFileDialog.getSaveFileName(
            self, "Export image", "", "Images (*.png)"
        )

        if not dst[0]:
            self.l.debug("No destination file, exiting func screenshot")
            return

        # Copy merged STL into destination file
        self.stl_viewer.screenshot(dst[0])

    def viewSTL(self, file_name: str):

        """
        This method is called by renderSTLs, and will update the stl viewer
        widget. file_name is the name of the file for the module to be
        loaded in the viewer
        """

        self.l.debug(f"Entering viewSTL with {file_name}")

        self.stl_viewer.viewMesh(file_name)

    def toggleTranslucent(self):

        """Toggle the transparency state of the STL viewer"""

        self.stl_viewer.toggleTranslucent()

        # Change button state
        if self.stl_viewer.translucent:
            self.button_translucent.setChecked(True)
        else:
            self.button_translucent.setChecked(False)

    def togglePerspective(self):

        """
        Toggle the perspective button. Set it to checked or unchecked. Toggle
        the perspective in the STL viewer
        """

        self.stl_viewer.togglePerspective()

        # Change button state
        if self.stl_viewer.perspective:
            self.button_perspective.setChecked(True)
        else:
            self.button_perspective.setChecked(False)

    def toggleLinearAssembly(self):

        """
        Toggle the values of x turn and y turn to equal 1 for a linear assembly. Set it to checked or unchecked.
        """
        self.ass.toggleLinearAssembly()

        # Change button state
        if self.ass.linear:
            self.button_linear.setChecked(True)
            self.ass.turn_x: int = 1
            self.ass.turn_y: int = 1
            self.refresh()
            print('Changed x turn and y turn to 1 i.e. linear')
        else:
            self.button_linear.setChecked(False)
            #self.ass.turn_x: int = 2
            #self.ass.turn_y: int = 2
            Settings().restoreParams()
            self.refresh()
            print('x and y turn values restore to default i.e. = 2')

    def exportSTL(self, stl_filename):

        """Slot called to export assembly as a STL"""

        self.l.debug("Trying to export assembly as STL")

        mes = "Before exporting the assembly as a STL, we first must render it"

        # Check the assembly is rendered
        if not self.rendered:
            self.l.debug("Assembly not rendered, display dailogbox and exit")
            QtWidgets.QMessageBox.warning(self, "Export STL", mes)
            return

        # First we will find the modules where visibility was
        # set to OFF from the checkbox. These won't be rendered
        meshes_ignore = []
        for key, value in self.stl_viewer.meshes.items():
            # if visibility was OFF, then alpha is 0
            # otherwise alpha would be 1
            if value.filter.alpha == 0:
                meshes_ignore.append(key)

        # Merge all the stl files of the modules
        self.ass.mergeSTLs(meshes_ignore)

        # Let user choose destination file
        #dst = QtWidgets.QFileDialog.getSaveFileName(
        #    self, "Export STL", "", "STL files (*.stl *.STL)"
        #)

        #if not dst[0]:
        #    self.l.debug("No destination file, exiting exportSTL")
        #    return

        src = os.path.join(self.renders_dir, "final.stl")

        # Copy merged STL into destination file
        shutil.copyfile(src, stl_filename)

        self.l.debug("Assembly exported as STL")
        self.l.debug(f"Saving project to {stl_filename}")

    def saveProject(self):

        """Save the project (current assembly) in a .ccad file"""

        self.l.debug("Saving project")

        # Let user choose saving file
        dst = QtWidgets.QFileDialog.getSaveFileName(
            self, "Save project", "", "CCAD files (*.ccad *.CCAD)"
        )

        if not dst[0]:
            self.l.debug("No destination file, exiting saveProject")
            return

        self.l.debug(f"Saving project to {dst}")

        # Save main attributes of assembly
        try:
            self.ass.saveProject(dst[0])
            stl_filename = "{}.stl".format(os.path.splitext(dst[0])[0])
            self.exportSTL(stl_filename)
        except Exception as e:
            self.l.error(f"Couldn't save project: {e}", exc_info=True)

        saved = QtWidgets.QMessageBox.information(self,
                                         "SAVED!",
                                         "CCAD and STL files have been saved successfully in the chosen directory",
                                         QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        if saved == QtWidgets.QMessageBox.Ok:
            return
        else:
            return



    def openProject(self):

        """Open a project, a pickled assembly stored as a .ccad file"""

        self.l.debug("Opening project")

        fnm = QtWidgets.QFileDialog.getOpenFileName(
            self, "Select a project", "", "CCAD files (*.ccad *.CCAD)"
        )

        if not fnm[0]:
            self.l.debug("openProject, no file selected, exiting")

        self.l.debug(f"openProject, trying to open {fnm[0]}")

        self.ass.openProject(fnm[0])

        self.l.debug(f"openProject, {fnm[0]} opened")

        self.refresh()

    def newProject(self):# creates new ChemCAD window - HPD - 25/11/19
        self.w = NewWindow()
        self.w.show()
        #self.hide()

    def clearLayout(self, layout: QtWidgets.QLayout):

        """Method to erase the widgets from a layout"""

        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                widget = item.widget()
                if widget is not None:
                    widget.deleteLater()

                    QtWidgets.qApp.processEvents()
                else:
                    self.clearLayout(item.layout())

    def initUI(self):

        """Build the interface"""

        self.setWindowTitle("ChemSCAD")

        # enable the status bar
        self.statusBar().showMessage('Ready')

        # font = QtGui.QFont()
        # font.setStyleHint(QtGui.QFont.System)
        # font.setPointSize(self.styles.FONT_SIZE)
        # font.setStyleStrategy(QtGui.QFont.PreferAntialias)
        # QtWidgets.qApp.setFont(font)

        # self.l.debug('Font: {}'.format(font.family()))
        # self.l.debug('Font size: {}pt'.format(self.styles.FONT_SIZE))

        # Empty widget acting like a spacer
        # empty_widget = QtWidgets.QWidget()
        # empty_widget.setSizePolicy(QtWidgets.QSizePolicy.Expanding,
        # QtWidgets.QSizePolicy.Preferred)

        # ------------------------- DASHBOARD ---------------------------

        # Add the custom widget Dashboard
        self.dashboard = Dashboard(self)

        # ------------------------- PREVIEW AREA ------------------------------

        # https://stackoverflow.com/questions/45964913/how-to-embed-vispy-canvas-in-pyqt5-frame
        # Create the STL viewer widget, with dimensions and watch folder
        # TODO: adapt dimensions depending on window's size
        self.stl_viewer = STLViewer((800, 500), self.renders_dir, self.l)

        # ------------------------- CONTROL CENTER ----------------------------

        # Will create attribute self.scroll_control_center
        self.buildControlCenter()

        # ------------------------- ASSEMBLING THE AREAS ----------------------

        self.split_right = QtWidgets.QSplitter(QtCore.Qt.Vertical)
        self.split_right.addWidget(self.stl_viewer.native)
        self.split_right.addWidget(self.scroll_control_center)

        self.split_main = QtWidgets.QSplitter(QtCore.Qt.Horizontal)
        self.split_main.addWidget(self.dashboard)
        self.split_main.addWidget(self.split_right)

        self.setCentralWidget(self.split_main)

        self.buildMenuBar()
        self.buildToolBar()

    def buildControlCenter(self):

        """
        Create the scroll area for the checkboxes triggering the display of each module
        """

        # Create the scroll area for the checkboxes triggering the display of
        # each module
        self.scroll_control_center = QtWidgets.QScrollArea()
        self.scroll_control_center.setStyleSheet("border: none;")
        self.scroll_control_center.setWidgetResizable(True)

        # Create a vbox layout to arrange the checkboxes vertically
        self.vbox_control_center = QtWidgets.QVBoxLayout()
        self.vbox_control_center.setAlignment(QtCore.Qt.AlignTop)

        # Assign an intermediate widget to the scroll are, and give it the vbox.
        # It doesn't make any sense to me but that's how items works
        self.scrolling_control_center = QtWidgets.QWidget()
        self.scrolling_control_center.setLayout(self.vbox_control_center)

    def buildMenuBar(self):

        """Build the Menu bar. Called by initUI"""

        self.menubar = self.menuBar()

        # Building files menu
        self.fileMenu = self.menubar.addMenu("&Files")
        # self.fileMenu.addAction(self.settingsAction)
        self.fileMenu.addAction(self.newProjectAction)
        self.fileMenu.addAction(self.openProjectAction)
        self.fileMenu.addAction(self.saveProjectAction)
        #self.fileMenu.addAction(self.exportSTLAction)
        self.fileMenu.addAction(self.renderProdAction)
        self.fileMenu.addAction(self.screenshotAction)
        self.fileMenu.addAction(self.preferencesAction)
        self.fileMenu.addAction(self.filterSettingsAction)
        self.fileMenu.addAction(self.openDatabaseAction)
        self.fileMenu.addAction(self.exitAction)

        # Building tools menu
        # self.toolMenu = self.menubar.addMenu("&Tools")
        # self.toolMenu.addAction(self.parseAction)

        # self.viewMenu = self.menubar.addMenu("&View")
        # self.sortMenu = self.viewMenu.addMenu("Sorting")
        # self.sortMenu.addAction(self.sortingPercentageAction)

        self.editMenu = self.menubar.addMenu("&Edit")
        self.editMenu.addAction(self.undoAction)
        self.editMenu.addAction(self.redoAction)

        self.helpMenu = self.menubar.addMenu("&Help")
        #self.helpMenu.addAction(self.aboutAction)
        self.helpMenu.addAction(self.helpAction)

    def buildToolBar(self):

        """Build the Tool bar. Called by initUI"""

        # ------------------------- TOOLBAR  ----------------------------------

        # Add a toolbar and name it to identiy it
        # Then add the widgets
        self.toolbar = self.addToolBar("toolbar")
        self.toolbar.setObjectName("toolbar")
        self.toolbar.setMovable(False)

        # "Add module" button, with an icon
        plus_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "plus.png"))
        self.button_add_module = QtWidgets.QPushButton(self)
        self.button_add_module.setIcon(plus_icon)
        self.button_add_module.setStatusTip("Add module")
        self.toolbar.addWidget(self.button_add_module)

        # "Align tops" button, with an icon
        align_tops_icon = QtGui.QIcon(
            os.path.join(self.resource_dir, "images", "align_tops.png")
        )
        self.button_align_tops = QtWidgets.QPushButton(self)
        self.button_align_tops.setIcon(align_tops_icon)
        self.button_align_tops.setStatusTip("Align tops")
        self.toolbar.addWidget(self.button_align_tops)

        # "Align filter" button, with an icon
        align_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "align.png"))
        self.button_align_filters = QtWidgets.QPushButton(self)
        self.button_align_filters.setIcon(align_icon)
        self.button_align_filters.setStatusTip("Align filter")
        self.toolbar.addWidget(self.button_align_filters)

        # Button render
        render_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "render.png"))
        self.button_render = QtWidgets.QPushButton("Render model")
        self.button_render = QtWidgets.QPushButton(self)
        self.button_render.setIcon(render_icon)
        self.button_render.setStatusTip("Render model")
        self.toolbar.addWidget(self.button_render)

        # Button for UHD render
        UHD_render_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "UHD-render.png"))
        self.button_renderUHD = QtWidgets.QPushButton("Render UHD model")
        self.button_renderUHD = QtWidgets.QPushButton(self)
        self.button_renderUHD.setIcon(UHD_render_icon)
        self.button_renderUHD.setStatusTip("Render UHD model")
        self.toolbar.addWidget(self.button_renderUHD)

        # Button undo
        undo_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "undo.png"))
        self.button_undo = QtWidgets.QPushButton("Undo action")
        self.button_undo = QtWidgets.QPushButton(self)
        self.button_undo.setIcon(undo_icon)
        self.button_undo.setStatusTip("Undo action")
        self.toolbar.addWidget(self.button_undo)


        # Button redo
        redo_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "redo.png"))
        self.button_redo = QtWidgets.QPushButton("Redo action")
        self.button_redo = QtWidgets.QPushButton(self)
        self.button_redo.setIcon(redo_icon)
        self.button_redo.setStatusTip("Redo action")
        self.toolbar.addWidget(self.button_redo)


        # Button translucent
        self.button_translucent = QtWidgets.QPushButton("X-ray")
        self.button_translucent.setCheckable(True)
        self.button_translucent.setStatusTip("X-ray render")
        self.toolbar.addWidget(self.button_translucent)

        # Button perspective
        self.button_perspective = QtWidgets.QPushButton("Perspective")
        self.button_perspective.setCheckable(True)
        self.button_perspective.setStatusTip("Reset prespective")
        self.toolbar.addWidget(self.button_perspective)

        # Button Save
        save_icon = QtGui.QIcon(os.path.join(self.resource_dir, "images", "save-icon.png"))
        self.button_save = QtWidgets.QPushButton("Save project")
        self.button_save = QtWidgets.QPushButton(self)
        self.button_save.setIcon(save_icon)
        self.button_save.setStatusTip("Save project")
        self.toolbar.addWidget(self.button_save)

        # Button linear
        self.button_linear = QtWidgets.QPushButton("Linear")
        self.button_linear.setCheckable(True)
        self.button_linear.setStatusTip("Linear assembly")
        self.toolbar.addWidget(self.button_linear)

class CommandBuild(QtWidgets.QUndoCommand):
    """
    In order to perform undo and redo from build commands
    """

    def __init__(self, cad: ObjectCAD, ass: Assembly, parent: QtWidgets.QMainWindow):

        super(CommandBuild, self).__init__()
        self.cad_object = cad
        self.ass = ass
        self.parent = parent

        # # if the object is connector we need to save IOs
        # if isinstance(self.cad_object, ConnectorCAD):
        #     self.in_io = self.cad_object.in_io
        #     self.out_io = self.cad_object.out_io

    def redo(self):
        # if the object is a connector
        # we need to refresh ios
        # if isinstance(self.cad_object, ConnectorCAD):
        #     self.cad_object.reconnectIOs(self.in_io, self.out_io)

        self.ass.appendModule(self.cad_object)
        self.parent.refresh()

    def undo(self):
        # delete module from assembly
        self.ass.deleteModule(self.cad_object)
        self.parent.refresh()
        # remove the mesh from the stl_viewer
        self.parent.stl_viewer.deleteMesh(self.cad_object.internal_id)
        self.parent.refresh()

class CommandDelete(QtWidgets.QUndoCommand):
    """
    In order to perform undo and redo from delete commands
    """

    def __init__(self, cad: ObjectCAD, ass: Assembly, parent: QtWidgets.QMainWindow):

        super(CommandDelete, self).__init__()
        self.cad_object = cad
        self.ass = ass
        self.parent = parent

        # # if the object is connector we need to save IOs
        if isinstance(self.cad_object, ConnectorCAD):
            self.in_io = self.cad_object.in_io
            self.out_io = self.cad_object.out_io

    def redo(self):
        # delete module from assembly
        self.ass.deleteModule(self.cad_object)
        self.parent.refresh()
        # remove the mesh from the stl_viewer
        self.parent.stl_viewer.deleteMesh(self.cad_object.internal_id)
        self.parent.refresh()

    def undo(self):
        # if the object is a connector
        # we need to refresh IOs
        if isinstance(self.cad_object, ConnectorCAD):
            self.cad_object.reconnectIOs(self.in_io, self.out_io)

        # Now we add back the reactor or connector
        self.ass.appendModule(self.cad_object)

        # if is a reactor, we need to add the connectors it had (if any)
        if isinstance(self.cad_object, ReactorCAD):
            for _, inp in self.cad_object.inputs.items():
                con_module = inp.connectedTo
                if con_module:
                    con_module.reconnectIOs(con_module._old_in_io,
                        con_module._old_out_io)
                    self.ass.appendModule(con_module)

            for _, outp in self.cad_object.outputs.items():
                con_module = outp.connectedTo
                if con_module:
                    con_module.reconnectIOs(con_module._old_in_io,
                        con_module._old_out_io)
                    self.ass.appendModule(con_module)

        self.parent.refresh()

class NewWindow(MyWindow): #HPD - 25/11/19
    def __init__(self):
        super().__init__()
        self.setWindowTitle("New ChemCAD Window")

    def closeEvent(self, event):
        close = QtWidgets.QMessageBox.question(self,
                                         "QUIT",
                                         "Are you sure want to close this project?",
                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if close == QtWidgets.QMessageBox.Yes:
            event.accept()
        else:
            event.ignore()



def main():
    # logger = MyLog()
    # try:
    app = QtWidgets.QApplication(sys.argv)

    # ex = Fenetre(logger)
    ex = MyWindow()
    app.processEvents()
    sys.exit(app.exec_())
    print("\n".join(repr(w) for w in app.allWidgets()))

if __name__ == "__main__":
    main()



#!/usr/bin/python
# coding: utf-8

import sys
from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Union, List, Dict, Optional
from ccad import ReactorCAD
from ccad.object import ObjectCAD

from chemscad.log import MyLog

class help_info(QtWidgets.QDialog):

    """
    Display an 'About ChemSCAD' window. The user can view info about the software including contributers, repository, current version, pip package info etc.
    """

    def __init__(self, parent=None):

        super(help_info, self).__init__(parent)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        if parent is None:
            self.l = MyLog("activity.log")
            self.test = True
        else:
            self.l = self.parent.l
            self.options = self.parent.options
            self.test = False

        self.initUI()
        self.defineSlots()

    def defineSlots(self):

        """Establish the slots"""

        # User finished selecting tops to align
        self.ok_button.clicked.connect(self.closeSettings)

    def closeSettings(self):

        """
        Close the settings window
        """

        if self.test:
            return

        # General refresh from main window
        self.parent.refresh()

        self.close()


    def initUI(self):
        self.title="About ChemSCAD"
        self.ok_button = QtWidgets.QPushButton("OK", self)
        msgBox = QtWidgets.QMessageBox()
        msgBox.setIcon(QtWidgets.QMessageBox.Information)
        msgBox.setText("Name: ChemSCAD\nVersion: 2.0\nDescription: software for generating reactionware objects")
        msgBox.setWindowTitle(self.title)
        msgBox.setStandardButtons(QtWidgets.QMessageBox.Ok | QtWidgets.QMessageBox.Cancel)
        msgBox.exec()


if __name__ == '__main__':
    app = QtWidgets.QApplication(sys.argv)
    obj = help_info()
    sys.exit(app.exec_())




#!/usr/bin/python
# coding: utf-8

import logging
from logging.handlers import RotatingFileHandler


class MyLog(logging.Logger):

    """Custom logger class. A pre-configured logger is created when the class
    is called. Writes into the terminal and in activity.log.
    Use: logger.(info|warn|debug|error|critical)
    http://sametmax.com/ecrire-des-logs-en-python/"""

    def __init__(self, output_file="activity.log", mode="a", total=True, size=5000000):

        super(MyLog, self).__init__(self)

        # Set the logging level to DEBUG
        self.setLevel(logging.DEBUG)

        # Set the format of te output
        if total:
            self.formatter = logging.Formatter(
                "%(asctime)s :: %(levelname)s :: %(message)s"
            )
        else:
            self.formatter = logging.Formatter("%(message)s")

        # Create a handler, redirects the output to a file in 'append' mode, w/
        # a backup and a mx size of 5 Mo
        self.file_handler = RotatingFileHandler(output_file, mode, size, 1)
        self.file_handler.setLevel(logging.DEBUG)
        self.file_handler.setFormatter(self.formatter)
        self.addHandler(self.file_handler)

        # Second handler to get the output into the terminal
        self.steam_handler = logging.StreamHandler()
        self.steam_handler.setLevel(logging.DEBUG)
        self.addHandler(self.steam_handler)


if __name__ == "__main__":
    log = MyLog()
    log.info("Hello")
    log.warning("Testing %s", "foo")



#!/usr/bin/python
# coding: utf-8

import pytest
import numpy as np
import itertools
import subprocess as sp
import trimesh
import os
import solid as solid

from ccad.log import MyLog
from ccad.tops_placement import get_positions
from ccad.top_in import TopInlet


l = MyLog("output_tests_top_placement.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


def test_get_positions():

    """
    Test the autoplacement of several top inlets. Check that the autoplacement
    function works for different numbers of inlets: 1, 2, 3, 4, 5
    """

    io1 = TopInlet(5, 7, 3, "custom", 0, 0)
    io2 = TopInlet(5, 7, 3, "custom", 0, 0)
    io3 = TopInlet(5, 7, 3, "custom", 0, 0)
    io4 = TopInlet(5, 7, 3, "custom", 0, 0)
    io5 = TopInlet(5, 7, 3, "custom", 0, 0)

    positions = get_positions(16, [io1])
    logAssert(positions == [(0, 0)], "Bad placement for 1 inlet")

    positions = get_positions(16, [io1, io2])
    logAssert(positions == [(-8.0, 0), (8.0, 0)], "Bad placement for 2 inlets")

    positions = get_positions(16, [io1, io2, io3])
    logAssert(
        positions
        == [
            (7.4256258422040737, 4.2871870788979622),
            (-7.4256258422040737, 4.2871870788979622),
            (-1.5750869800554844e-15, -8.5743741577959263),
        ],
        "Bad placement for 3 inlets",
    )

    positions = get_positions(16, [io1, io2, io3, io4])
    logAssert(
        positions
        == [
            (6.6274169979695206, 6.6274169979695188),
            (-6.6274169979695188, 6.6274169979695206),
            (-6.6274169979695214, -6.6274169979695188),
            (6.6274169979695179, -6.6274169979695214),
        ],
        "Bad placement for 4 inlets",
    )

    positions = get_positions(16, [io1, io2, io3, io4, io5])
    logAssert(
        positions
        == [
            (5.9230705305400022, 8.1524071919108607),
            (-5.9230705305400013, 8.1524071919108607),
            (-9.583729436176597, -3.1139424571808605),
            (-1.8511019130011733e-15, -10.076929469459998),
            (9.5837294361765952, -3.113942457180864),
        ],
        "Bad placement for 5 inlets",
    )


if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

import pytest
import os
import sys
import shutil

from ccad.log import MyLog
from ccad.exceptions import *
from ccad.assembly import Assembly
from ccad.reactor import ReactorCAD as rea
from ccad.siphon import SiphonCAD as siph


l = MyLog("output_tests_floating_filter_reactor.log", mode="w")

l.debug("---------------------- START NEW RUN OF TESTS ----------------------")
l.debug("\n")


def logAssert(test, msg):

    """
    Function to log the result of an assert
    http://stackoverflow.com/questions/24892396/py-test-logging-messages-and-test-results-assertions-into-a-single-file
    """

    if not test:
        l.error(msg)
        assert test, msg


@pytest.fixture()
def dummy_ass():

    """Dummy function for demo/test, create an assembly with 9 reactors"""

    ass = Assembly()

    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    r_2 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_2.addInputPercentage("input_test")
    ass.appendModule(r_2)

    con1 = siph(r_1, r_1.outputs["default"], r_2, r_2.inputs["input_test"])
    ass.appendModule(con1)

    r_3 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_3.addInputPercentage("input_test")
    ass.appendModule(r_3)

    con2 = siph(r_2, r_2.outputs["default"], r_3, r_3.inputs["input_test"])
    ass.appendModule(con2)

    r_4 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_4.addInputPercentage("input_test")
    ass.appendModule(r_4)

    con3 = siph(r_3, r_3.outputs["default"], r_4, r_4.inputs["input_test"])
    ass.appendModule(con3)

    r_5 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_5.addInputPercentage("input_test")
    ass.appendModule(r_5)

    con4 = siph(r_4, r_4.outputs["default"], r_5, r_5.inputs["input_test"])
    ass.appendModule(con4)

    r_6 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_6.addInputPercentage("input_test")
    ass.appendModule(r_6)

    con5 = siph(r_5, r_5.outputs["default"], r_6, r_6.inputs["input_test"])
    ass.appendModule(con5)

    r_7 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_7.addInputPercentage("input_test")
    ass.appendModule(r_7)

    con6 = siph(r_6, r_6.outputs["default"], r_7, r_7.inputs["input_test"])
    ass.appendModule(con6)

    r_8 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_8.addInputPercentage("input_test")
    ass.appendModule(r_8)

    con7 = siph(r_7, r_7.outputs["default"], r_8, r_8.inputs["input_test"])
    ass.appendModule(con7)

    r_9 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_9.addInputPercentage("input_test")
    ass.appendModule(r_9)

    con8 = siph(r_8, r_8.outputs["default"], r_9, r_9.inputs["input_test"])
    ass.appendModule(con8)

    ass.refresh()

    # ass.saveProject("Dummy.ccad")
    # print(ass)
    # ass.renderToFile("dummy_assembly.scad")
    # print(ass)

    return ass


def test_reactor_name_property(dummy_ass):

    l.debug("Testing assembly reactor names")

    ass = dummy_ass

    # Create target lists to compare to
    comparison = ["R{}".format(x) for x in range(1, 10)]

    # Get names for reactors in assembly
    names = ass.reactors_names.keys()

    for name, comp in zip(names, comparison):
        logAssert(name == comp, "Name {} does not match!".format(name))

    # Change name of first reactor
    ass.list_reactors[0].module_name = "Reactor test"

    names = list(ass.reactors_names.keys())

    logAssert(names[0] == "Reactor test", "Name does not match after renaming reactor")


def test_reactor_internal_id_property(dummy_ass):

    l.debug("Testing assembly reactor id")

    ass = dummy_ass

    # Create target lists to compare to
    comparison = ["R{}".format(x) for x in range(1, 10)]

    # Get IDs for reactors in assembly
    internal_ids = ass.reactors_internal_ids.keys()

    for internal_id, comp in zip(internal_ids, comparison):
        logAssert(internal_id == comp, "ID {} does not match!".format(internal_id))

    # Change name of first reactor
    ass.list_reactors[0].module_name = "Reactor test"

    internal_ids = list(ass.reactors_internal_ids.keys())

    logAssert(internal_ids[0] == "R1", "ID does not match after renaming reactor")


def test_connector_name_property(dummy_ass):

    l.debug("Testing assembly connector names")

    ass = dummy_ass

    # Create target lists to compare to
    comparison = ["S{}".format(x) for x in range(1, 9)]

    # Get names for connectors in assembly
    names = ass.connectors_names.keys()

    for name, comp in zip(names, comparison):
        logAssert(name == comp, "Connector {} does not match!".format(name))

    # Change name of first reactor
    ass.list_connectors[0].module_name = "Connector test"

    names = list(ass.connectors_names.keys())

    logAssert(
        names[0] == "Connector test", "Name does not match after renaming connector"
    )


def test_connector_internal_id_property(dummy_ass):

    l.debug("Testing assembly connector IDs")

    ass = dummy_ass

    # Create target lists to compare to
    comparison = ["S{}".format(x) for x in range(1, 9)]

    # Get names for connectors in assembly
    internal_ids = ass.connectors_internal_ids.keys()

    for internal_id, comp in zip(internal_ids, comparison):
        logAssert(
            internal_id == comp, "Connector {} does not match!".format(internal_id)
        )

    # Change name of first reactor
    ass.list_connectors[0].module_name = "Connector test"

    internal_ids = list(ass.connectors_internal_ids.keys())

    logAssert(internal_ids[0] == "S1", "ID does not match after renaming connector")


def test_get_module_id_from_name(dummy_ass):

    l.debug("Testing getting ID from module's name")

    ass = dummy_ass

    # Change name of first reactor
    ass.list_reactors[0].module_name = "Reactor test"

    logAssert(ass.getModuleIDFromName("Reactor test") == "R1", "Can't get ID from name")


def test_io_naming_properties(dummy_ass):

    l.debug("Testing Input and Output naming properties")

    ass = dummy_ass

    inputs = ["Input 'input_test' from R{}".format(x) for x in range(2, 10)]
    outputs = ["Output 'default' from R{}".format(x) for x in range(1, 10)]

    for test, target in zip(inputs, ass.inputs.keys()):
        logAssert(test == target, "{} is not present!".format(test))

    for test, target in zip(outputs, ass.outputs.keys()):
        logAssert(test == target, "{} is not present!".format(test))


def test_openscad_path(dummy_ass):

    l.debug("Testing openscad path")

    ass = dummy_ass

    if sys.platform == "linux":
        logAssert(ass.openscad_path == "openscad", "Openscad paths do not match!!")
    elif sys.platform == "osx" or sys.platform == "macosx":
        with pytest.raises(NotImplementedError):
            ass.openscad_path
    elif sys.platform == "win32":
        logAssert(
            "/ccad/ccad/bin/win/openscad." in ass.openscad_path,
            "Windows OpenSCAD path does not match!",
        )


def test_merge_stl(dummy_ass):

    l.debug("Testing merging of STLs")

    ass = dummy_ass
    ass.renders_dir = os.path.join("tests", "test_renders")
    ass.mergeSTLs()

    # final_target is the reference, compare the current rendering to it
    final_test = os.path.join("tests", "test_renders", "final_target.stl")

    # Forge path for the stl file just rendered
    final = os.path.join("tests", "test_renders", "final.stl")

    l.debug("Comparing test file to generated file")
    with open(final, "rb") as f1:
        with open(final_test, "rb") as f2:

            # Remove file being compared
            os.remove(final)

            content1 = f1.readlines()
            content2 = f2.readlines()

            logAssert(content1 == content2, "STL files do not match!")

    l.debug("Comparison passes")


def test_export_to_dict(dummy_ass):

    dico = dummy_ass.exportAssemblyToDict()

    logAssert(dico["turn_x"] == 2, "Wrong turn_x after export")
    logAssert(dico["turn_y"] == 2, "Wrong turn_y after export")

    logAssert(len(dico["list_reactors"]) == 9, "Wrong nbr of reactors after export")
    logAssert(len(dico["list_connectors"]) == 8, "Wrong nbr of reactors after export")

    logAssert(not dico["list_align_tops"], "List align tops: pb after export")
    logAssert(not dico["list_align_filters"], "List align filter: pb after export")


def test_save_project(dummy_ass):

    l.debug("Testing saving project")

    ass = dummy_ass
    ass.saveProject("test.ccad")

    proper_save = os.path.join("tests", "test_renders", "proper_save.ccad")

    with open(proper_save, "rb") as f1:
        content1 = f1.readlines()

    with open("test.ccad", "rb") as f2:
        os.remove("test.ccad")
        content2 = f2.readlines()

    logAssert(content1 == content2, "Saved Projects do not match!")


def test_load_project(dummy_ass):

    l.debug("Testing loading project")

    ass = dummy_ass

    proper_save = os.path.join("tests", "test_renders", "proper_save.ccad")
    test_ass = Assembly()
    test_ass.openProject(proper_save)

    logAssert(ass.turn_x == test_ass.turn_x, "Turn X does not match!")
    logAssert(ass.turn_y == test_ass.turn_y, "Turn Y does not match!")

    for test_r, target_r in zip(test_ass.list_reactors, ass.list_reactors):
        logAssert(test_r.infos == target_r.infos, "Reactors do not match!")

    for test_c, target_c in zip(test_ass.list_connectors, ass.list_connectors):
        logAssert(test_c.infos == target_c.infos, "Connectors do not match!")

    for test_at, target_at in zip(test_ass.list_align_tops, ass.list_align_tops):
        logAssert(test_at == target_at, "Align tops do not match!")

    for test_af, target_af in zip(test_ass.list_align_filters, ass.list_align_filters):
        logAssert(test_af == target_af, "Align filters do not match!")


def test_obj_must_be_rendered():

    l.debug("Testing object render checks")

    ass = Assembly()
    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)

    r_2 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_2.addInputPercentage("input_test")

    ass.appendModule(r_1)
    logAssert(ass._objMustBeRendered(r_1) is True, "Render Error, returned False")

    ass.appendModule(r_2)
    logAssert(ass._objMustBeRendered(r_2) is False, "Render Error, returned True")

    con1 = siph(r_1, r_1.outputs["default"], r_2, r_2.inputs["input_test"])
    ass.appendModule(con1)
    logAssert(ass._objMustBeRendered(r_1) is True, "render error, returned false")
    logAssert(ass._objMustBeRendered(r_2) is True, "Render error, returned false")


def test_obj_is_new():

    l.debug("Testing object is new")

    ass = Assembly()
    ass.renders_dir = os.path.join("tests", "test_renders")

    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    logAssert(ass._objIsNew(r_1) is True, "Obj is new but comparison failed")


def test_obj_has_changed():

    l.debug("Testing object has changed")

    ass = Assembly()
    ass.renders_dir = os.path.join("tests", "test_renders")

    # R1_has_changed is a scad file for a 20 mL reactor. Duplicate it as R1.scad
    # file. Mimics the process of rendering the object
    src = os.path.join(ass.renders_dir, "R1_has_changed.scad")
    dst = os.path.join(ass.renders_dir, "R1.scad")

    shutil.copyfile(src, dst)

    # Create a 21 mL reactor, to be compared to the 20 mL reactor in the scad file
    r_1 = rea(21, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    logAssert(
        ass._objHasChanged(r_1) is True, "Obj must be rendered but comparison failed"
    )

    # Clean render dir, remove R1.scad to not mess up other tests
    os.remove(dst)


def test_find_first_reactor_in_chain():

    l.debug("Testing assembly can find 1st reactor in chain")

    ass = Assembly()

    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    r_2 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_2.addInputPercentage("input_test")
    ass.appendModule(r_2)

    con1 = siph(r_1, r_1.outputs["default"], r_2, r_2.inputs["input_test"])
    ass.appendModule(con1)

    r_3 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_3.addInputPercentage("input_test")
    ass.appendModule(r_3)

    con2 = siph(r_2, r_2.outputs["default"], r_3, r_3.inputs["input_test"])
    ass.appendModule(con2)

    r_4 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_4.addInputPercentage("input_test")
    ass.appendModule(r_4)

    con3 = siph(r_3, r_3.outputs["default"], r_4, r_4.inputs["input_test"])
    ass.appendModule(con3)

    ass.refresh()

    # ass.renderToFile("test_first.scad")

    logAssert(
        ass._findFirstReactorInChain() is r_1,
        "Assembly didn't find the right 1st reactor in chain",
    )


def test_find_next_connector():

    l.debug("Testing assembly can find next connector")

    ass = Assembly()

    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    r_2 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_2.addInputPercentage("input_test")
    ass.appendModule(r_2)

    con1 = siph(r_1, r_1.outputs["default"], r_2, r_2.inputs["input_test"])
    ass.appendModule(con1)

    r_3 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_3.addInputPercentage("input_test")
    ass.appendModule(r_3)

    con2 = siph(r_2, r_2.outputs["default"], r_3, r_3.inputs["input_test"])
    ass.appendModule(con2)

    r_4 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_4.addInputPercentage("input_test")
    ass.appendModule(r_4)

    con3 = siph(r_3, r_3.outputs["default"], r_4, r_4.inputs["input_test"])
    ass.appendModule(con3)

    ass.refresh()

    # ass.renderToFile("test_first.scad")

    logAssert(
        ass._findNextConnector(r_2) is con2,
        "Assembly didn't find the right next connector for r_2",
    )


def test_reorder_connectors():

    l.debug("Testing assembly can reorder connectors")

    ass = Assembly()

    r_1 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    ass.appendModule(r_1)

    r_2 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_2.addInputPercentage("input_test")
    ass.appendModule(r_2)

    r_3 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_3.addInputPercentage("input_test")
    ass.appendModule(r_3)

    r_4 = rea(20, rea.T_OPEN, rea.B_ROUND_IN_O)
    r_4.addInputPercentage("input_test")
    ass.appendModule(r_4)

    con1 = siph(r_2, r_2.outputs["default"], r_3, r_3.inputs["input_test"])
    ass.appendModule(con1)

    con2 = siph(r_1, r_1.outputs["default"], r_2, r_2.inputs["input_test"])
    ass.appendModule(con2)

    con3 = siph(r_3, r_3.outputs["default"], r_4, r_4.inputs["input_test"])
    ass.appendModule(con3)

    ass._reorderConnectors()

    logAssert(
        ass.list_connectors == [con2, con1, con3],
        "Assembly didn't reorder connectors properly",
    )


if __name__ == "__main__":
    dummy_ass()



#!/usr/bin/python
# coding: utf-8

"""
Tutorials on how to create a lib:
http://sametmax.com/creer-un-setup-py-et-mettre-sa-bibliotheque-python-en-ligne-sur-pypi/
http://peterdowns.com/posts/first-time-with-pypi.html
https://github.com/SolidCode/SolidPython/blob/master/setup.py
"""

from setuptools import setup, find_packages

with open("README.md") as fd:
    readme = fd.read()

with open("requirements.txt") as fd:
    install_requires = [l for l in fd.readlines()]

setup(
    name="chemscad",
    version="2.0.2.b2",
    description="GUI application for the creation of Reactionware.",
    long_description=readme,
    long_description_content_type="text/markdown",
    author="Cronin Group",
    author_email="croningp.pypi@gmail.com",
    url="https://github.com/croningp/chemscad",
    include_package_data=True,
    packages=find_packages(),
    classifiers=[
        'Programming Language :: Python :: 3',
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        'Operating System :: OS Independent'
    ],
    install_requires=install_requires,
    python_requires=">=3.6",
    entry_points={
        "console_scripts": [
            "chemscad=chemscad.main:main"
        ]
    }
)



#!/usr/bin/python
# coding: utf-8


import sys
import os
from typing import Union
from PyQt5 import QtGui, QtCore, QtWidgets

from ccad import TubeConnectorCAD as tube
from ccad.exceptions import *


class ChemModule(QtWidgets.QDialog):

    """Tube connector module. To connect reactor-like objects"""

    def __init__(self, parent=None, cad_obj: Union[None, tube] = None):

        """Dialog box for tube"""

        super(ChemModule, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        self.cad_obj = cad_obj

        # Dict to store module's features
        self.params = dict()

        if parent is None:
            sys.path.append(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
            from log import MyLog

            self.l = MyLog("activity.log")

            # # Dummy file for saving if testing
            # self.options = QtCore.QSettings("debug/options.ini",
            # QtCore.QSettings.IniFormat)

            self.test = True
        else:
            self.l = self.parent.l
            # self.options = self.parent.options
            self.test = False

        self.path = os.path.dirname(os.path.abspath(__file__))
        self.l.debug(f"Module path: {self.path}")

        self.initUI()
        self.defineSlots()

        # Dicts to store I/O and objects by name
        self.dict_inputs = dict()
        self.dict_outputs = dict()
        self.dict_objects = dict()

        # Load the values in combo boxes and fill dicts above
        self.loadValues()
        self.adjustSize()  # Adjust size of dialog to fit contents of combo boxes.
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # To close the window and build the module
        self.ok_button.clicked.connect(self.buildModule)

        # To close the window and delete the module
        self.del_button.clicked.connect(self.deleteModule)
        self.del_button.setShortcut("Del")

    def readForm(self):

        """
        Read the form, and get the parameters of the module. Put them in a dict
        """

        # Get the name of the input I/O
        in_io_name = self.combo_in_io.currentText()
        out_io_name = self.combo_out_io.currentText()

        # If one I/O has no name, it was empty. Display error message and exit
        if not in_io_name or not out_io_name:
            self.l.debug("Missing I/O for tube")

            mes = "At least one I/O is missing"
            QtWidgets.QMessageBox.critical(
                self, "Missing IO", mes, QtWidgets.QMessageBox.Ok
            )

            raise MissingConnector("Missing I/O")

        self.params["in_io"] = self.dict_outputs[in_io_name]
        self.params["out_io"] = self.dict_inputs[out_io_name]

        # Get the input object from the name of input I/O.
        # Avoids using another combo box for input object
        obj_in_nbr = in_io_name[-1]
        for name in self.dict_objects:
            if str(obj_in_nbr) in name:
                obj_in = self.dict_objects[name]
                break

        # Get the output object from the name of output I/O.
        # Avoids using another combo box for output object
        obj_out_nbr = out_io_name[-1]
        for name in self.dict_objects:
            if obj_out_nbr in name:
                obj_out = self.dict_objects[name]
                break

        self.params["obj_in"] = obj_in
        self.params["obj_out"] = obj_out

        # Get the text for the conflicts strategy
        conflicts = self.combo_conflicts.currentText()

        # Transform the text to an API str parameter
        if conflicts == "Lift input object":
            self.params["conflicts"] = "lift_in_obj"
        elif conflicts == "Lift output object":
            self.params["conflicts"] = "lift_out_obj"
        elif conflicts == "Move output":
            self.params["conflicts"] = "move_out_io"

        length = self.spin_length.value()
        self.params["length"] = length

        self.l.debug(f"tube, form read: {self.params}")

    def restoreParams(self):

        """
        If a CAD object is provided, we are updating the object.
        Restore the parameters previously entered to the display
        """

        # If no cad_obj, we're creating a module, nothing to restore
        if self.cad_obj is None:
            return

        # Restore input and output objects to parameters
        for name, obj in self.dict_objects.items():
            if obj is self.cad_obj.obj_in:
                self.params["obj_in"] = obj
                continue
            elif obj is self.cad_obj.obj_out:
                self.params["obj_out"] = obj
                continue

        # Restore input to parameters
        for name, io in self.dict_outputs.items():
            if io is self.cad_obj.in_io:
                self.params["in_io"] = io
                self.combo_in_io.setCurrentIndex(self.combo_in_io.findText(name))
                break

        # Restore output to parameters
        for name, io in self.dict_inputs.items():
            if io is self.cad_obj.out_io:
                self.params["out_io"] = io
                self.combo_out_io.setCurrentIndex(self.combo_out_io.findText(name))
                break

        # Restore the conflict strategy in the combo box
        if self.cad_obj.conflicts == "lift_in_obj":
            self.combo_conflicts.setCurrentText("Lift input object")
        elif self.cad_obj.conflicts == "lift_out_obj":
            self.combo_conflicts.setCurrentText("Lift output object")
        elif self.cad_obj.conflicts == "move_out_io":
            self.combo_conflicts.setCurrentText("Move output")

        # Restore length of connector
        self.params["length"] = self.cad_obj.length
        self.spin_length.setValue(self.params["length"])
        self.spin_length.setEnabled(True)

        self.l.debug(f"TubeConnector, parameters restored: {self.params}")

    def buildModule(self):

        """
        Called when user clicks the OK button. Will read the form, build the 3D
        model of the module, and tell the parent to display the module
        """

        # TODO: generate an image of the module, and build the CAD object

        # Don't do anything if missing I/O
        try:
            self.readForm()
        except MissingConnector:
            self.l.error("Siphon called w missing connector. Exit")
            return

        if not self.test:
            if self.cad_obj is None:
                self._buildNewModule()
            else:
                self._updateModule()

        # Close the settings window and free the memory
        self.close()

    def deleteModule(self):

        """
        Called when user clicks the delete button. 
        It will delete the selected module
        """

        self.l.debug("Tube, deleteModule")
        self.parent.deleteModule(self.cad_obj)

        # Close the settings window and free the memory
        self.close()

    def _buildNewModule(self) -> bool:

        """
        Internal method called when building the object for the first time
        """

        self.l.debug("TubeConnector, _buildNewModule")

        obj_in = self.params["obj_in"]
        in_io = self.params["in_io"]
        obj_out = self.params["obj_out"]
        out_io = self.params["out_io"]
        conflicts = self.params["conflicts"]

        try:
            con = tube(obj_in, in_io, obj_out, out_io, conflicts)
        except Exception as e:
            mes = f"Impossible to create tube connector: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Creation error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        self.parent.buildModule(con)

        return True

    def _updateModule(self) -> bool:

        self.l.debug("TubeConnector, _updateModule")

        obj_in = self.params["obj_in"]
        in_io = self.params["in_io"]
        obj_out = self.params["obj_out"]
        out_io = self.params["out_io"]
        conflicts = self.params["conflicts"]
        length = self.params["length"]

        self.cad_obj.obj_in = obj_in
        self.cad_obj.in_io = in_io
        self.cad_obj.obj_out = obj_out
        self.cad_obj.out_io = out_io
        self.cad_obj.conflicts = conflicts
        self.cad_obj.length = length

        # TODO: check if object changed before calling refresh
        self.parent.refresh()

        return True

    def loadValues(self):

        """
        Load the values/name for combo boxes (choices for I/O and objects)
        Only consider unconnected I/O, disable others
        https://stackoverflow.com/questions/29971030/pyqt-enable-disable-elements-in-a-qcombobox
        """

        self.l.debug("Siphon, loading values")

        if self.test:
            return

        # List of objects
        self.dict_objects = self.parent.ass.reactors_names

        self.dict_outputs = self.parent.ass.outputs
        for name, io in self.dict_outputs.items():
            self.combo_in_io.addItem(name)

            # Disable I/O in choices if connected
            if io.connected:
                index = self.combo_in_io.count() - 1
                self.combo_in_io.model().item(index).setEnabled(False)

        self.dict_inputs = self.parent.ass.inputs
        for name, io in self.dict_inputs.items():
            self.combo_out_io.addItem(name)

            # Disable I/O in choices if connected
            if io.connected:
                index = self.combo_out_io.count() - 1
                self.combo_out_io.model().item(index).setEnabled(False)

    def initUI(self):

        """Handles the display"""

        self.setWindowTitle("Add a tube connector")

        # self.tabs = QtWidgets.QTabWidget(self)

        # ----------------- BASIC SETTINGS ------------------------------------

        fbox = QtWidgets.QFormLayout()
        # Combo box to choose input I/O
        self.combo_in_io = QtWidgets.QComboBox(self)
        self.combo_in_io.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        # Combo box to choose output I/O
        self.combo_out_io = QtWidgets.QComboBox(self)
        self.combo_out_io.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)

        self.combo_conflicts = QtWidgets.QComboBox(self)
        self.combo_conflicts.setSizeAdjustPolicy(QtWidgets.QComboBox.AdjustToContents)
        list_conflicts = ["Lift input object", "Lift output object", "Move output"]
        self.combo_conflicts.addItems(list_conflicts)

        # Combo box to set length of connector
        self.spin_length = QtWidgets.QDoubleSpinBox(self)
        self.spin_length.setMinimum(0)
        self.spin_length.setEnabled(False)

        # Add widgets to form layout
        fbox.addRow("Input I/O: ", self.combo_in_io)
        fbox.addRow("Output I/O: ", self.combo_out_io)
        fbox.addRow("Conflicts: ", self.combo_conflicts)
        fbox.addRow("Length: ", self.spin_length)

        # ----------------- ASSEMBLING ----------------------------------------

        self.ok_button = QtWidgets.QPushButton("OK", self)
        self.del_button = QtWidgets.QPushButton("Delete", self)
        split_butttons = QtWidgets.QHBoxLayout()
        split_butttons.addWidget(self.ok_button)
        split_butttons.addWidget(self.del_button)

        # if new object only display ok button
        if self.cad_obj is None:
            self.del_button.setVisible(False)

        fbox.addRow(split_butttons)

        self.setLayout(fbox)
        self.show()


if __name__ == "__main__":
    print(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    app = QtWidgets.QApplication(sys.argv)
    obj = ChemModule()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8

import os, math, sys
from typing import Union
from PyQt5 import QtGui, QtCore, QtWidgets
from time import sleep
import time as time

from ccad import FlowReactorCAD as rea
from add_io import AddIODialog
from add_top_io import AddTopIODialog
from ccad.exceptions import *
import ccad.constants as cst


class ChemModule(QtWidgets.QDialog):

    """Flow Reactor module. """

    def __init__(self, parent=None, cad_obj: Union[None, rea] = None):

        super(ChemModule, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        self.cad_obj = cad_obj

        if parent is None:
            sys.path.append(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
            from log import MyLog

            self.l = MyLog("activity.log")

            # # Dummy file for saving if testing
            # self.options = QtCore.QSettings("debug/options.ini",
            # QtCore.QSettings.IniFormat)

            self.test = True
        else:
            self.l = self.parent.l
            # self.options = self.parent.options
            self.test = False

        self.path = os.path.dirname(os.path.abspath(__file__))
        self.l.debug(f"Module path: {self.path}")

        self.initUI()
        self.defineSlots()

        # Store inputs
        self.dict_inputs = dict()

        # Store outputs
        self.dict_outputs = dict()

        # Store top inlets
        self.dict_top_inlets = dict()

        # Dict to store module's features
        self.params = dict()
        self.restoreParams()

        self.volume = self.calculateVolume()

    def defineSlots(self):

        """Establish the slots"""

        # To close the window and build the module
        #self.ok_button.clicked.connect(self.buildModule)

        # To close the window and delete the module
        self.del_button.clicked.connect(self.deleteModule)
        self.del_button.setShortcut("Del")

        # Update volume when spinboxes change
        self.spin_npipes.valueChanged.connect(self.calculateVolume)
        self.spin_width.valueChanged.connect(self.calculateVolume)
        self.spin_radius.valueChanged.connect(self.calculateVolume)
        self.spin_spacing.valueChanged.connect(self.calculateVolume)
        self.spin_ninputs.valueChanged.connect(self.calculateVolume)


    def calculateVolume(self):

        """ Calculates the volume of the internal channels """

        npipes = self.spin_npipes.value()
        width = self.spin_width.value()
        radius = self.spin_radius.value()
        spacing = self.spin_spacing.value()
        inputs = self.spin_ninputs.value()
        
        total_volume, flowpath = self.getVolume(npipes, width, radius, spacing, inputs)

        text = "{:.{}f}".format( total_volume/1000, 2 )
        self.vol_label.setText(text +" ml")

        text = "{:.{}f}".format( flowpath/1000, 2 )
        self.flowpath_label.setText(text +" ml")


    def getVolume(self, npipes, width, radius, spacing, inputs):

        # First calculate the volume of the horizontal pipes
        vhpipe = math.pi * width * radius**2 # volume of 1 horizontal pipe
        vhpipes = vhpipe * npipes # multiply by npies

        # now volume of curved connecting pipes
        # volume of curved pipe is volume of half torus
        torus1 = math.pi*radius**2
        torus2 = 2*math.pi*spacing/2
        torus = (torus1*torus2) / 2 # we only have half torus
        ncurved = npipes + 1
        volcurved = torus*ncurved

        # each in/out has a quarter of a torus, remember torus is /2
        toio = torus
        # last 2 horizontal (half sized) pipes connecting to input output
        last_horizontal = math.pi*(radius**2)*(width/2-spacing/2)
        toio += last_horizontal * 2
        # we add the last horizontal pipe bit connecting to outlet
        toio += math.pi*(radius**2)*spacing

        # now the volume of the inlets
        invol = math.pi*(radius**2)*(spacing/2)
        if inputs in [1,3]: # add volume of central outlet
            invol += math.pi*(radius**2)*spacing
        if inputs in [2,3]: # add volume of 2 extra outlets
            invol += math.pi*(radius**2)*20 # horizontal pipe - HPD - 13/02/20 - changed 10 to 18 to reflect change in distpipes value
            invol += torus # 2 last 1/4 torus
            invol += (math.pi*(radius**2)*spacing)*2#last piping

        total_volume = vhpipes + volcurved + toio + invol
        flowpath = vhpipes + volcurved + toio

        return total_volume, flowpath

    def restoreParams(self):

        """
        If a CAD object is provided, we are updating the object.
        Restore the parameters previously entered to the display
        """

        # If no cad_obj, we're creating a module, nothing to restore
        if self.cad_obj is None:
            return

        # Get the values of the object from cad object
        self.params["volume"] = self.cad_obj.volume
        self.params["npipes"] = self.cad_obj.npipes
        self.params["ninputs"] = self.cad_obj.ninputs
        self.params["width"] = self.cad_obj.width
        self.params["radius"] = self.cad_obj.radius
        self.params["spacing"] = self.cad_obj.spacing

        # Set values for graphical elements
        self.spin_npipes.setValue(self.params["npipes"])
        self.spin_ninputs.setValue(self.params["ninputs"])
        self.spin_width.setValue(self.params["width"])
        self.spin_radius.setValue(self.params["radius"])
        self.spin_spacing.setValue(self.params["spacing"])
        self.spin_volume.setValue(self.params["volume"])

        self.restoreIO()

        self.l.debug(f"Flow Reactor, restored flow reactor: {self.params}")

    def restoreIO(self):

        """Restore the I/O of the cad object into the dialog box"""

        # Restore all inputs for object
        for name, in_io in self.cad_obj.inputs.items():
            height_per = self.cad_obj.getHeightPercentage(in_io)
            angle = in_io.angle
            diameter = in_io.diameter
            external = in_io.external
            connected = in_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
            }

            self.l.debug(f"Flow Reactor, restoring I/O {params}")

            self.dict_inputs[name] = params

            # Add I/O to the table
            self._addNewRow(name, self.SIDE_INPUT, connected)

        # Restore all outputs for object
        for name, out_io in self.cad_obj.outputs.items():
            height_per = self.cad_obj.getHeightPercentage(out_io)
            angle = out_io.angle
            diameter = out_io.diameter
            external = out_io.external
            connected = out_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
            }

            self.l.debug(f"Flow Reactor, restoring I/O {params}")

            self.dict_outputs[name] = params

            # Add I/O to the table
            if name == "default":
                # Different type of output if default
                self._addNewRow(name, self.DEF_OUTPUT, connected)
            else:
                self._addNewRow(name, self.SIDE_OUTPUT, connected)

        self.l.debug("Flow Reactor, restored I/O")

    def readForm_Basic(self):

        """
        Read the form, and get the parameters of the module. Put them in a dict 
        """
        self.params["npipes"] = self.spin_npipes.value()
        self.params["ninputs"] = self.spin_ninputs.value()
        self.params["width"] = self.spin_width.value()
        self.params["radius"] = self.spin_radius.value()
        self.params["spacing"] = self.spin_spacing.value()
        self.params["volume"] = self.spin_volume.value()
        self.l.debug(f"Flow reactor, read form {self.params}")

    def readForm_Advanced(self):

        """
        Read the form, and get the parameters of the module. Put them in a dict 
        """
        self.params["npipes"] = self.spin_npipes.value()
        self.params["ninputs"] = self.spin_ninputs.value()
        self.params["width"] = self.spin_width.value()
        self.params["radius"] = self.spin_radius.value()
        self.params["spacing"] = self.spin_spacing.value()
        self.params["volume"] = self.spin_volume.value()
        self.l.debug(f"Flow reactor, read form {self.params}")

    def setTemplate(self, template=None):

        template_1 = {
            'npipes': 8,
            'ninputs': 2,
            'width': 20,
            'radius': 0.8,
            'spacing': 6,
            'volume' : 1
            }
        template_2 = {
            'npipes': 24,
            'ninputs': 2,
            'width': 20,
            'radius': 0.8,
            'spacing': 6,
            'volume' : 1
            }
        template_3 = {
            'npipes': 8,
            'ninputs': 3,
            'width': 20,
            'radius': 0.8,
            'spacing': 6,
            'volume' : 1
            }
        template_4 = {
            'npipes': 24,
            'ninputs': 3,
            'width': 20,
            'radius': 0.8,
            'spacing': 6,
            'volume' : 1
            }

        if template == 'Template 1':
            for key, value in template_1.items():
                self.params[key] = value
                self.l.debug(f"Flow reactor, read form {self.params}")
            if not self.test:
            # If no CAD object provided, we're creating a new object
                if self.cad_obj is None:
                    self._buildNewModule()
                else:
                    self._updateModule()

        if template == 'Template 2':
            for key, value in template_2.items():
                self.params[key] = value
                self.l.debug(f"Flow reactor, read form {self.params}")
            if not self.test:
            # If no CAD object provided, we're creating a new object
                if self.cad_obj is None:
                    self._buildNewModule()
                else:
                    self._updateModule()

        if template == 'Template 3':
            for key, value in template_3.items():
                self.params[key] = value
                self.l.debug(f"Flow reactor, read form {self.params}")
            if not self.test:
            # If no CAD object provided, we're creating a new object
                if self.cad_obj is None:
                    self._buildNewModule()
                else:
                    self._updateModule()


        if template == 'Template 4':
            for key, value in template_4.items():
                self.params[key] = value
                self.l.debug(f"Flow reactor, read form {self.params}")
            if not self.test:
            # If no CAD object provided, we're creating a new object
                if self.cad_obj is None:
                    self._buildNewModule()
                else:
                    self._updateModule()
            # Close the settings window and free the memory


    def addInput(self, params: dict):

        """The AddIODialog class will call this method"""

        self.l.debug(f"Flow Reactor, adding input: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_inputs, add it to the table
        if name not in self.dict_inputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_reactor: addInput {params}")
        else:
            self.l.debug(f"addInput, updating input {name} with {params}")

        self.dict_inputs[name] = params

    def addOutput(self, params: dict):

        """The AddIODialog class will call this method"""

        self.l.debug(f"Flow Reactor, adding output: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_outputs, add it to the table
        if name not in self.dict_outputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_reactor: addOutput {params}")
        else:
            self.l.debug(f"addOutput, updating output {name} with {params}")

        self.dict_outputs[name] = params

    def _addNewRow(self, name: str, type_io: str, connected: bool = False):

        """Add a new row in the I/O table"""

        self.l.debug(f"Adding new row {name}, {type_io}")

        # Count nbr of rows
        row_count = self.table_io.rowCount()

        # Insert row at the end
        self.table_io.insertRow(row_count)

        # Create name and type of I/O items to insert in the table
        item_name = QtWidgets.QTableWidgetItem(name)

        # TopInlet objects have an attribute 'type_io'. InputOutput objects
        # don't, their type is determined based on which dict they are in
        # (inputs or outputs)
        if type_io == "custom":
            item_type_io = QtWidgets.QTableWidgetItem("Custom top inlet")
        elif type_io == "luer":
            item_type_io = QtWidgets.QTableWidgetItem("Luer top inlet")
        else:
            item_type_io = QtWidgets.QTableWidgetItem(type_io)

        # Create an item depending on the connection state of the I/O
        if connected:
            item_connected = QtWidgets.QTableWidgetItem("Connected")
        else:
            item_connected = QtWidgets.QTableWidgetItem("Not connected")

        # Disable selection for default output; can't do anything on it
        # from I/O tab
        if type_io == self.DEF_OUTPUT:
            item_name.setFlags(QtCore.Qt.NoItemFlags)
            item_type_io.setFlags(QtCore.Qt.NoItemFlags)
            item_connected.setFlags(QtCore.Qt.NoItemFlags)

        # Insert the 2 items in the table
        self.table_io.setItem(row_count, 0, item_name)
        self.table_io.setItem(row_count, 1, item_type_io)
        self.table_io.setItem(row_count, 2, item_connected)

    def buildInputs(self, cad_obj):

        """Build an input for the CAD object"""

        self.l.debug("Reactor, building inputs")

        for name, in_io in self.dict_inputs.items():
            cad_obj.addInputPercentage(
                name,
                in_io["height_per"],
                in_io["angle"],
                in_io["diameter"],
                in_io["external"],
            )

    def buildOutputs(self, cad_obj):

        """Build an ouput for the CAD object"""

        self.l.debug("Reactor, building outputs")

        for name, out_io in self.dict_outputs.items():

            # Skip default output
            if name == "default":
                continue

            cad_obj.addOutputPercentage(
                name,
                out_io["height_per"],
                out_io["angle"],
                out_io["diameter"],
                out_io["external"],
            )

    def buildModule_Advanced(self):

        """
        Called when user clicks the OK button. Will read the form, build the 3D
        model of the module, and tell the parent to display the module
        """

        # TODO: generate an image of the module, and build the CAD object

        self.readForm_Advanced()

        if not self.test:
            # If no CAD object provided, we're creating a new object
            if self.cad_obj is None:
                self._buildNewModule()
            else:
                self._updateModule()

        # Close the settings window and free the memory
        self.close()

    def buildModule_Basic(self):

        """
        Called when user clicks the OK button. Will read the form, build the 3D
        model of the module, and tell the parent to display the module
        """

        # TODO: generate an image of the module, and build the CAD object

        self.readForm_Basic()

        if not self.test:
            # If no CAD object provided, we're creating a new object
            if self.cad_obj is None:
                self._buildNewModule()
            else:
                self._updateModule()

        # Close the settings window and free the memory
        self.close()

    def deleteModule(self):

        """
        Called when user clicks the delete button.
        It will delete the selected module
        """

        self.l.debug("Flow Reactor, deleteModule")
        self.parent.deleteModule(self.cad_obj)

        # Close the settings window and free the memory
        self.close()

    def _buildNewModule(self) -> bool:

        """
        Internal method called when building the object for the first time
        """

        self.l.debug("Flow Reactor, _buildNewModule")

        try:
            reactor = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to create flow reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Creation error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Try to build I/Os
        try:
            self.buildInputs(reactor)
            self.buildOutputs(reactor)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        self.parent.buildModule(reactor)

        return True

    def _updateModule(self) -> bool:

        """
        Internal method called when updating the object with new form values
        """

        self.l.debug("Flow Reactor, _updateModule")

        # Try to build a temporary reactor with the new parameters
        # If it fails, exit immediately
        try:
            _ = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to update flow reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Update error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # If temporary reactor was built, update cad_object
        self.l.debug("Flow Reactor, buildModule w cad_obj")

        self.cad_obj.npipes = self.params["npipes"]
        self.cad_obj.ninputs = self.params["ninputs"]
        self.cad_obj.width = self.params["width"]
        self.cad_obj.radius = self.params["radius"]
        self.cad_obj.spacing = self.params["spacing"]
        self.cad_obj.volume = self.params["volume"]

        # Try to build I/Os
        try:
            self.buildInputs(self.cad_obj)
            self.buildOutputs(self.cad_obj)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Try to refresh assembly
        try:
            self.parent.refresh()
        except Exception as e:
            mes = f"Impossible to refresh flow reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Refresh error", mes, QtWidgets.QMessageBox.Ok
            )

        return True

    def _buildCADObject(self) -> rea:

        """
        Use the parameters read from the form to build the CAD object. Does
        nothing else, no GUI task here
        """

        npipes = self.params["npipes"]
        ninputs = self.params["ninputs"]
        width = self.params["width"]
        radius = self.params["radius"]
        spacing = self.params["spacing"]
        volume = self.params["volume"]

        reactor = rea(
            npipes,
            ninputs,
            width,
            radius,
            spacing,
            volume,
        )

        return reactor

    def changeTemplate(self, event):
        changeTemplate = QtWidgets.QMessageBox.question(self,
                                         "Change Template",
                                         "Would you like to change the selected flow reactor parameters? Go to ADVANCED",
                                         QtWidgets.QMessageBox.Yes | QtWidgets.QMessageBox.No)
        if changeTemplate == QtWidgets.QMessageBox.Yes:
            pass
        else:
            self.close()


    def buttonClicked(self):
        sender = self.sender()
        if sender == self.template1_button:
            self.setTemplate(template='Template 1')
            #self.changeTemplate(event)
        elif sender == self.template2_button:
            self.setTemplate(template='Template 2')
            #self.changeTemplate(event)
        elif sender == self.template3_button:
            self.setTemplate(template='Template 3')
            #self.changeTemplate(event)
        elif sender == self.template4_button:
            self.setTemplate(template='Template 4')
            #self.changeTemplate(event)
        self.close()

    def basicUpdated(self):

        """ Update basic tab values """

        npipes = 10
        width = 10
        spacing = 6
        inputs = 3
        radius = self.spin_radius.value() # spin_radius_basic
        volume = self.spin_volume.value() # if volume = 10, width = some value

        
        total_volume, flowpath = self.getVolume(npipes, width, radius, spacing, inputs)

        text = "{:.{}f}".format( total_volume/1000, 2 )
        self.vol_label.setText(text +" ml")

        text = "{:.{}f}".format( flowpath/1000, 2 )
        self.flowpath_label.setText(text +" ml")
        print("Radius was updated...")

    def initUI(self):

        """Handles the display"""

        if self.cad_obj is None:
            self.setWindowTitle("Add a flow reactor module")
        else:
            self.setWindowTitle("Modify flow reactor module")

        # Build the basic settings tab
        widget_form_basic = self._buildBasicSettings()

        # Build the advanced settings tab
        widget_form_advanced = self._buildAdvancedSettings()

        # Build the templates tab
        widget_form_template = self._buildTemplateSettings()

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)

        self.tabs = QtWidgets.QTabWidget(self)

        self.tabs.addTab(widget_form_basic, "Basic")
        self.tabs.addTab(widget_form_template, "Templates")
        self.tabs.addTab(widget_form_advanced, "Advanced")


        
        self.del_button = QtWidgets.QPushButton("Delete", self)
        split_butttons = QtWidgets.QHBoxLayout()
        #split_butttons.addWidget(self.ok_button)
        split_butttons.addWidget(self.del_button)

        # if new object only display ok button
        if self.cad_obj is None:
            self.del_button.setVisible(False)

        # self.vbox_global.addLayout(fbox)
        self.vbox_global.addWidget(self.tabs)
        self.vbox_global.addLayout(split_butttons)

        self.setLayout(self.vbox_global)
        self.show()



    def _buildBasicSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the basic settings"""

        widget_form_basic = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_basic.setLayout(fbox)

        # Spin box to specify radius pipes
        self.spin_radius = QtWidgets.QDoubleSpinBox(self)
        # Set default and bound values
        self.spin_radius.setValue(1)
        self.spin_radius.setMinimum(0.1)
        self.spin_radius.setMaximum(200)
        self.spin_radius.valueChanged.connect(self.basicUpdated)

        # Spin box to specify volume of flow path
        self.spin_volume = QtWidgets.QDoubleSpinBox(self)
        # Set default and bound values
        self.spin_volume.setValue(1)
        self.spin_volume.setMinimum(0.1)
        self.spin_volume.setMaximum(100)
        self.spin_volume.valueChanged.connect(self.basicUpdated)


        # text label for total volume
        self.vol_label = QtWidgets.QLabel("0", self)

        # text label for flowpath volume
        self.flowpath_label = QtWidgets.QLabel("0", self)

        # Add widgets to form layout
        fbox.addRow("Volume (mL) of flow path: ", self.spin_volume)
        fbox.addRow("Radius (mm) of all the pipes: ", self.spin_radius)
        fbox.addRow("Total volume of reactor: ", self.vol_label)
        fbox.addRow("Volume of flow path: ", self.flowpath_label)

        self.ok_button_basic = QtWidgets.QPushButton("OK", self)
        self.ok_button_basic.clicked.connect(self.buildModule_Basic)
        fbox.addRow("", self.ok_button_basic)

        return widget_form_basic

    def _buildAdvancedSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the template settings"""

        widget_form_advanced = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_advanced.setLayout(fbox)

        # Spin box to specify number of horizontal pipes
        self.spin_npipes = QtWidgets.QSpinBox(self)
        # Set default and bound values
        self.spin_npipes.setValue(4)
        self.spin_npipes.setMinimum(1)
        self.spin_npipes.setMaximum(200)

        # Spin box to specify number of inlets
        self.spin_ninputs = QtWidgets.QSpinBox(self)
        # Set default and bound values
        self.spin_ninputs.setValue(2)
        self.spin_ninputs.setMinimum(1)
        self.spin_ninputs.setMaximum(3)

        # Spin box to specify width of horizontal pipes
        self.spin_width = QtWidgets.QDoubleSpinBox(self)
        # Set default and bound values
        self.spin_width.setValue(10)
        self.spin_width.setMinimum(1)
        self.spin_width.setMaximum(200)

        # Spin box to specify radius pipes
        self.spin_radius = QtWidgets.QDoubleSpinBox(self)
        # Set default and bound values
        self.spin_radius.setValue(1)
        self.spin_radius.setMinimum(0.1)
        self.spin_radius.setMaximum(200)

        # Spin box to specify spacing between pipes
        self.spin_spacing = QtWidgets.QDoubleSpinBox(self)
        # Set default and bound values
        self.spin_spacing.setValue(6)
        self.spin_spacing.setMinimum(1)
        self.spin_spacing.setMaximum(200)

        # text label for total volume
        self.vol_label = QtWidgets.QLabel("0", self)

        # text label for flowpath volume
        self.flowpath_label = QtWidgets.QLabel("0", self)

        # Add widgets to form layout
        fbox.addRow("Number horizontal pipes: ", self.spin_npipes)
        fbox.addRow("Number inputs: ", self.spin_ninputs)
        fbox.addRow("Width (mm) h-tal pipes: ", self.spin_width)
        fbox.addRow("Radius (mm) of all the pipes: ", self.spin_radius)
        fbox.addRow("Spacing (mm) between h-tal pipes: ", self.spin_spacing)
        fbox.addRow("Total volume of reactor: ", self.vol_label)
        fbox.addRow("Volume of flow path: ", self.flowpath_label)
        self.ok_button_advanced = QtWidgets.QPushButton("OK", self)
        fbox.addRow("", self.ok_button_advanced)
        self.ok_button_advanced.clicked.connect(self.buildModule_Advanced)

        return widget_form_advanced

    def _buildTemplateSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the template settings"""
        widget_form_template = QtWidgets.QWidget(self)
        fbox = QtWidgets.QVBoxLayout(self)
        widget_form_template.setLayout(fbox)

        self.template1_button = QtWidgets.QPushButton("Inputs: 2, Volume of Flow Path: 0.55 mL", self)
        self.template2_button = QtWidgets.QPushButton("Inputs: 2, Volume of Flow Path: 1.50 mL", self)
        self.template3_button = QtWidgets.QPushButton("Inputs: 3, Volume of Flow Path: 0.55 mL", self)
        self.template4_button = QtWidgets.QPushButton("Inputs: 3, Volume of Flow Path: 1.50 mL", self)

        fbox.addWidget(self.template1_button)
        fbox.addWidget(self.template2_button)
        fbox.addWidget(self.template3_button)
        fbox.addWidget(self.template4_button)
        self.template1_button.clicked.connect(self.buttonClicked)
        self.template2_button.clicked.connect(self.buttonClicked)
        self.template3_button.clicked.connect(self.buttonClicked)
        self.template4_button.clicked.connect(self.buttonClicked)

        #self.show()

        # # Spin box to specify number of horizontal pipes
        # self.spin_npipes = QtWidgets.QSpinBox(self)
        # # Set default and bound values
        # self.spin_npipes.setValue(4)
        # self.spin_npipes.setMinimum(1)
        # self.spin_npipes.setMaximum(200)

        # # Spin box to specify number of inlets
        # self.spin_ninputs = QtWidgets.QSpinBox(self)
        # # Set default and bound values
        # self.spin_ninputs.setValue(2)
        # self.spin_ninputs.setMinimum(1)
        # self.spin_ninputs.setMaximum(3)

        # # Spin box to specify width of horizontal pipes
        # self.spin_width = QtWidgets.QDoubleSpinBox(self)
        # # Set default and bound values
        # self.spin_width.setValue(10)
        # self.spin_width.setMinimum(1)
        # self.spin_width.setMaximum(200)

        # # Spin box to specify radius pipes
        # self.spin_radius = QtWidgets.QDoubleSpinBox(self)
        # # Set default and bound values
        # self.spin_radius.setValue(1)
        # self.spin_radius.setMinimum(0.1)
        # self.spin_radius.setMaximum(200)

        # # Spin box to specify spacing between pipes
        # self.spin_spacing = QtWidgets.QDoubleSpinBox(self)
        # # Set default and bound values
        # self.spin_spacing.setValue(6)
        # self.spin_spacing.setMinimum(1)
        # self.spin_spacing.setMaximum(200)

        # # text label for total volume
        # self.vol_label = QtWidgets.QLabel("0", self)

        # # text label for flowpath volume
        # self.flowpath_label = QtWidgets.QLabel("0", self)

        # # Add widgets to form layout
        # fbox.addRow("Number horizontal pipes: ", self.spin_npipes)
        # fbox.addRow("Number inputs: ", self.spin_ninputs)
        # fbox.addRow("Width (mm) h-tal pipes: ", self.spin_width)
        # fbox.addRow("Radius (mm) of all the pipes: ", self.spin_radius)
        # fbox.addRow("Spacing (mm) between h-tal pipes: ", self.spin_spacing)
        # fbox.addRow("Total volume of reactor: ", self.vol_label)
        # fbox.addRow("Volume of flow path: ", self.flowpath_label)

        return widget_form_template




if __name__ == "__main__":
    print(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    app = QtWidgets.QApplication(sys.argv)
    obj = ChemModule()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8

import sys
import os
from typing import Union
from PyQt5 import QtGui, QtCore, QtWidgets

from ccad import ReactorCAD as rea
from add_io import AddIODialog
from add_top_io import AddTopIODialog
from ccad.exceptions import *
import ccad.constants as cst


class ChemModule(QtWidgets.QDialog):

    """Reactor module. Simple cartridge where a reaction could happen"""

    # Define str for type of I/O
    SIDE_INPUT = "Side input"
    SIDE_OUTPUT = "Side output"
    DEF_OUTPUT = "Default output"
    TOP_LUER = "Luer top inlet"
    TOP_CUSTOM = "Custom top inlet"

    def __init__(self, parent=None, cad_obj: Union[None, rea] = None):

        super(ChemModule, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        self.cad_obj = cad_obj

        if parent is None:
            sys.path.append(
                os.path.dirname(
                    os.path.dirname(os.path.dirname(os.path.abspath(__file__)))
                )
            )
            from log import MyLog

            self.l = MyLog("activity.log")

            # # Dummy file for saving if testing
            # self.options = QtCore.QSettings("debug/options.ini",
            # QtCore.QSettings.IniFormat)

            self.test = True
        else:
            self.l = self.parent.l
            # self.options = self.parent.options
            self.test = False

        self.path = os.path.dirname(os.path.abspath(__file__))
        self.l.debug(f"Module path: {self.path}")

        self.initUI()
        self.defineSlots()

        # Store inputs
        self.dict_inputs = dict()

        # Store outputs
        self.dict_outputs = dict()

        # Store top inlets
        self.dict_top_inlets = dict()

        # Dict to store module's features
        self.params = dict()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # Bottom combo box changed
        self.combo_bottom.currentIndexChanged.connect(self.bottomTypeChanged)

        # Top combo box changed
        self.combo_top.currentIndexChanged.connect(self.topTypeChanged)

        # Radius constraint checkbox changed
        self.check_r_constrained.stateChanged.connect(self.rConstraintChanged)

        # To close the window and build the module
        self.ok_button.clicked.connect(self.buildModule)

        # To close the window and delete the module
        self.del_button.clicked.connect(self.deleteModule)
        self.del_button.setShortcut("Del")

        # Open dialog box to add I/O
        self.button_add_io.clicked.connect(self.openTypeIODialog)

        # Open dialog box to modify I/O from list of inputs
        self.table_io.itemDoubleClicked.connect(self.openIODialog)

        # Delete an I/O when button clicked
        self.button_delete_io.clicked.connect(self.deleteIO)

    def bottomTypeChanged(self):

        """
        Slot called when combo box for bottom changes. Will disable the
        spin box for diameter of internal pipe when bottom has no pipe
        """

        bottom_type = self.combo_bottom.currentText()

        if bottom_type in rea.BOTTOMS_PIPE:
            self.spin_d_can.setEnabled(True)
        else:
            self.spin_d_can.setEnabled(False)

    def topTypeChanged(self):

        """Disable the r spin box if top is threaded"""

        top_type = self.combo_top.currentText()

        if top_type in rea.THREADED_TOPS:
            # Threaded top:
            # - r_constrained checkbox is checked, do this first bc
            # checking it calls rConstraintChanged and enables spin radius
            # - disable setting radius
            # - disable check box for setting constraint on radius
            self.check_r_constrained.setCheckState(QtCore.Qt.Checked)
            self.spin_r.setEnabled(False)
            self.check_r_constrained.setEnabled(False)
        else:
            self.check_r_constrained.setEnabled(True)

    def rConstraintChanged(self):

        """
        Slot called when the state of the check box for the constraint of
        the radius changes. If it's not checked, radius is not constrained,
        disable spin box for radius. If it's checked, enable spin box
        for radius
        """

        if self.check_r_constrained.checkState() == QtCore.Qt.Unchecked:
            self.spin_r.setEnabled(False)
        elif self.check_r_constrained.checkState() == QtCore.Qt.Checked:
            self.spin_r.setEnabled(True)

    def deleteIO(self):

        """Slot: delete an input from the reactor"""

        # Get row for selected I/O
        try:
            row = self.table_io.selectionModel().selectedRows()[0].row()
        except IndexError:
            self.l.debug("No I/O selected, can't delete")
            return

        # Get name and type for I/O
        name = self.table_io.item(row, 0).text()
        type_io = self.table_io.item(row, 1).text()

        if type_io == self.DEF_OUTPUT:
            # We should never be here since selection is disabled for def
            # output, but just in case, return
            return

        elif type_io == self.SIDE_INPUT:
            del self.dict_inputs[name]
            # Delete I/O from CAD object
            if self.cad_obj is not None:
                del self.cad_obj.inputs[name]

        elif type_io == self.SIDE_OUTPUT:
            del self.dict_outputs[name]
            if self.cad_obj is not None:
                del self.cad_obj.outputs[name]

        else:
            del self.dict_top_inlets[name]
            if self.cad_obj is not None:
                del self.cad_obj.top_inlets[name]

        self.l.debug(f"Reactor, deleting I/O {name}")

        # Remove I/O line from table
        self.table_io.removeRow(row)

    def openTypeIODialog(self):

        """
        Create a small dialog box to let the user choose what type of I/O
        he wants to create: side I/O, or top I/O
        """

        dial = QtWidgets.QDialog(self)

        label_choice = QtWidgets.QLabel("What would you like to create:")

        button_side_io = QtWidgets.QPushButton("A side I/O")
        button_top_io = QtWidgets.QPushButton("A top I/O")

        button_side_io.clicked.connect(dial.accept)
        button_side_io.clicked.connect(lambda: AddIODialog(self))

        button_top_io.clicked.connect(dial.accept)
        button_top_io.clicked.connect(lambda: AddTopIODialog(self))

        hbox_dial = QtWidgets.QHBoxLayout()
        hbox_dial.addWidget(button_side_io)
        hbox_dial.addWidget(button_top_io)

        vbox_dial = QtWidgets.QVBoxLayout()
        vbox_dial.addWidget(label_choice, alignment=QtCore.Qt.AlignHCenter)
        vbox_dial.addLayout(hbox_dial)

        dial.setLayout(vbox_dial)

        dial.show()

    def openIODialog(self, item: QtWidgets.QTableWidgetItem):

        """
        Slot: open I/O dialog box when an I/O is double-clicked in the list
        of I/O
        """

        self.l.debug("Reactor, opening I/O dialog box")

        # Get row for double-clicked I/O
        row = item.row()

        # Get name and type for I/O
        name = self.table_io.item(row, 0).text()
        type_io = self.table_io.item(row, 1).text()

        # Create a dict with name and type of I/O
        params = {"name": name, "type_io": type_io}

        # Get parameters for I/O from dicts
        if type_io == self.SIDE_INPUT:
            # Merge simple dict params with full dict for I/O
            params = {**params, **self.dict_inputs[name]}
            AddIODialog(self, params)
        elif type_io == self.SIDE_OUTPUT:
            params = {**params, **self.dict_outputs[name]}
            AddIODialog(self, params)
        else:
            params = {**params, **self.dict_top_inlets[name]}
            AddTopIODialog(self, params)

    def restoreParams(self):

        """
        If a CAD object is provided, we are updating the object.
        Restore the parameters previously entered to the display
        """

        # If no cad_obj, we're creating a module, nothing to restore
        if self.cad_obj is None:
            return

        # Get the values of the object from cad object
        self.params["volume"] = self.cad_obj.volume
        self.params["type_bottom"] = self.cad_obj.type_bottom
        self.params["type_top"] = self.cad_obj.type_top

        # Advanced settings
        self.params["d_can"] = self.cad_obj.d_can
        self.params["r"] = self.cad_obj.r
        self.params["r_constrained"] = self.cad_obj.r_constrained

        # Set values for graphical elements
        self.spin_vol.setValue(self.params["volume"])
        self.combo_top.setCurrentIndex(rea.TOPS.index(self.params["type_top"]))
        self.combo_bottom.setCurrentIndex(rea.BOTTOMS.index(self.params["type_bottom"]))

        # Advanced settings
        self.spin_d_can.setValue(self.params["d_can"])
        self.spin_r.setValue(self.params["r"])

        # Restore check box state for r constrained
        if self.params["r_constrained"]:

            # Check check box for constraint on radius
            self.check_r_constrained.setCheckState(QtCore.Qt.Checked)
            if self.params["type_top"] in rea.THREADED_TOPS:
                # Threaded top:
                # - disable spin radius
                # - disable check box for constraint
                self.spin_r.setEnabled(False)
                self.check_r_constrained.setEnabled(False)
            else:
                self.check_r_constrained.setEnabled(True)

        else:
            self.check_r_constrained.setCheckState(QtCore.Qt.Unchecked)
            self.spin_r.setEnabled(False)

        # Restore the align top strategy
        align_top_strategy = self.cad_obj.align_top_strategy
        if align_top_strategy == "expand":
            self.combo_align_top_stratgey.setCurrentText("Expand body")
        elif align_top_strategy == "lift":
            self.combo_align_top_stratgey.setCurrentText("Lift reactor")

        self.restoreIO()

        self.l.debug(f"Reactor, restored reactor: {self.params}")

    def restoreIO(self):

        """Restore the I/O of the cad object into the dialog box"""

        # Restore all inputs for object
        for name, in_io in self.cad_obj.inputs.items():
            height_per = self.cad_obj.getHeightPercentage(in_io)
            angle = in_io.angle
            diameter = in_io.diameter
            external = in_io.external
            connected = in_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
            }

            self.l.debug(f"Reactor, restoring I/O {params}")

            self.dict_inputs[name] = params

            # Add I/O to the table
            self._addNewRow(name, self.SIDE_INPUT, connected)

        # Restore all outputs for object
        for name, out_io in self.cad_obj.outputs.items():
            height_per = self.cad_obj.getHeightPercentage(out_io)
            angle = out_io.angle
            diameter = out_io.diameter
            external = out_io.external
            connected = out_io.connected

            params = {
                "height_per": height_per,
                "angle": angle,
                "diameter": diameter,
                "external": external,
                "connected": connected,
            }

            self.l.debug(f"Reactor, restoring I/O {params}")

            self.dict_outputs[name] = params

            # Add I/O to the table
            if name == "default":
                # Different type of output if default
                self._addNewRow(name, self.DEF_OUTPUT, connected)
            else:
                self._addNewRow(name, self.SIDE_OUTPUT, connected)

        # Restore all top inlets for object
        for name, top_io in self.cad_obj.top_inlets.items():
            diameter = top_io.diameter
            length = top_io.length
            walls = top_io.walls
            type_io = top_io.type_io
            connected = False

            params = {
                "diameter": diameter,
                "length": length,
                "walls": walls,
                "type_io": type_io,
                "connected": connected,
            }

            self.l.debug(f"Reactor, restoring I/O {params}")

            self.dict_top_inlets[name] = params

            # Add I/O to the table
            self._addNewRow(name, type_io, connected)

        self.l.debug("Reactor, restored I/O")

    def readForm(self):

        """
        Read the form, and get the parameters of the module. Put them in a dict
        """

        self.params["volume"] = self.spin_vol.value()
        self.params["type_bottom"] = self.combo_bottom.currentText()
        self.params["type_top"] = self.combo_top.currentText()

        # Advanced settings
        self.params["d_can"] = self.spin_d_can.value()
        self.params["r"] = self.spin_r.value()

        # Check if r is constrained
        state = self.check_r_constrained.checkState()
        if state == QtCore.Qt.Unchecked:
            self.params["r_constrained"] = False
        elif state == QtCore.Qt.Checked:
            self.params["r_constrained"] = True

        # Get align top strategy and transform to usable string value
        strategy = self.combo_align_top_stratgey.currentText()
        if strategy == "Expand body":
            strategy = "expand"
        elif strategy == "Lift reactor":
            strategy = "lift"
        self.params["align_top_strategy"] = strategy

        self.l.debug(f"Reactor, read form {self.params}")

    def addInput(self, params: dict):

        """The AddIODialog class will call this method"""

        self.l.debug(f"Reactor, adding input: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_inputs, add it to the table
        if name not in self.dict_inputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_reactor: addInput {params}")
        else:
            self.l.debug(f"addInput, updating input {name} with {params}")

        self.dict_inputs[name] = params

    def addOutput(self, params: dict):

        """The AddIODialog class will call this method"""

        self.l.debug(f"Reactor, adding output: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_outputs, add it to the table
        if name not in self.dict_outputs:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_reactor: addOutput {params}")
        else:
            self.l.debug(f"addOutput, updating output {name} with {params}")

        self.dict_outputs[name] = params

    def addTopInlet(self, params: dict):

        """
        The AddTopIODialog class will call this method. Create a top inlet
        """

        self.l.debug(f"Reactor, adding top inlet: {params}")

        name = params["name"]
        del params["name"]

        # If input not already in dict_top_inlets, add it to the table
        if name not in self.dict_top_inlets:
            self._addNewRow(name, params["type_io"])
            self.l.debug(f"module_reactor: top inlet {params}")
        else:
            self.l.debug(f"addTopInlet, updating inlet {name} with {params}")

        self.dict_top_inlets[name] = params

    def _addNewRow(self, name: str, type_io: str, connected: bool = False):

        """Add a new row in the I/O table"""

        self.l.debug(f"Adding new row {name}, {type_io}")

        # Count nbr of rows
        row_count = self.table_io.rowCount()

        # Insert row at the end
        self.table_io.insertRow(row_count)

        # Create name and type of I/O items to insert in the table
        item_name = QtWidgets.QTableWidgetItem(name)

        # TopInlet objects have an attribute 'type_io'. InputOutput objects
        # don't, their type is determined based on which dict they are in
        # (inputs or outputs)
        if type_io == "custom":
            item_type_io = QtWidgets.QTableWidgetItem("Custom top inlet")
        elif type_io == "luer":
            item_type_io = QtWidgets.QTableWidgetItem("Luer top inlet")
        else:
            item_type_io = QtWidgets.QTableWidgetItem(type_io)

        # Create an item depending on the connection state of the I/O
        if connected:
            item_connected = QtWidgets.QTableWidgetItem("Connected")
        else:
            item_connected = QtWidgets.QTableWidgetItem("Not connected")

        # Disable selection for default output; can't do anything on it
        # from I/O tab
        if type_io == self.DEF_OUTPUT:
            item_name.setFlags(QtCore.Qt.NoItemFlags)
            item_type_io.setFlags(QtCore.Qt.NoItemFlags)
            item_connected.setFlags(QtCore.Qt.NoItemFlags)

        # Insert the 2 items in the table
        self.table_io.setItem(row_count, 0, item_name)
        self.table_io.setItem(row_count, 1, item_type_io)
        self.table_io.setItem(row_count, 2, item_connected)

    def buildInputs(self, cad_obj):

        """Build an input for the CAD object"""

        self.l.debug("Reactor, building inputs")

        for name, in_io in self.dict_inputs.items():
            cad_obj.addInputPercentage(
                name,
                in_io["height_per"],
                in_io["angle"],
                in_io["diameter"],
                in_io["external"],
            )

    def buildOutputs(self, cad_obj):

        """Build an ouput for the CAD object"""

        self.l.debug("Reactor, building outputs")

        for name, out_io in self.dict_outputs.items():

            # Skip default output
            if name == "default":
                continue

            cad_obj.addOutputPercentage(
                name,
                out_io["height_per"],
                out_io["angle"],
                out_io["diameter"],
                out_io["external"],
            )

    def buildTopInlets(self, cad_obj):

        """
        Build the top inlets. Also try to place them. For now, only
        auto-placement is supported
        """

        self.l.debug("Reactor, building top inlets")

        for name, io in self.dict_top_inlets.items():

            try:
                cad_obj.addCustomTopInlet(
                    name, io["diameter"], io["length"], io["walls"], io["type_io"]
                )
            except IncompatibilityError:
                self.l.debug("can't build top inlet: no custom top")
                return

        # TODO: handle manual placement (later)
        try:
            cad_obj.autoPlaceTopInlets()
        except ConstraintError:
            self.l.debug("Can't auto-place top inlets, collision ?")
            mes = """
            Impossible to auto-place the top inlets. Collision ?
            """

            QtWidgets.QMessageBox.critical(
                self, "Auto-placement error", mes, QtWidgets.QMessageBox.Ok
            )

            return

    def buildModule(self):

        """
        Called when user clicks the OK button. Will read the form, build the 3D
        model of the module, and tell the parent to display the module
        """

        # TODO: generate an image of the module, and build the CAD object

        self.readForm()

        if not self.test:
            # If no CAD object provided, we're creating a new object
            if self.cad_obj is None:
                self._buildNewModule()
            else:
                self._updateModule()

        # Close the settings window and free the memory
        self.close()

    def deleteModule(self):

        """
        Called when user clicks the delete button. 
        It will delete the selected module
        """

        self.l.debug("Reactor, deleteModule")
        self.parent.deleteModule(self.cad_obj)

        # Close the settings window and free the memory
        self.close()

    def _buildNewModule(self) -> bool:

        """
        Internal method called when building the object for the first time
        """

        self.l.debug("Reactor, _buildNewModule")

        try:
            reactor = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to create reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Creation error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Try to build I/Os
        try:
            self.buildInputs(reactor)
            self.buildOutputs(reactor)
            self.buildTopInlets(reactor)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        self.parent.buildModule(reactor)

        return True

    def _updateModule(self) -> bool:

        """
        Internal method called when updating the object with new form values
        """

        self.l.debug("Reactor, _updateModule")

        # Try to build a temporary reactor with the new parameters
        # If it fails, exit immediately
        try:
            _ = self._buildCADObject()
        except Exception as e:
            mes = f"Impossible to update reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Update error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Update radius and constraint on radius before updating rest
        # of the parameters, so if user uncheck the constraint check box,
        # reactor can be recalculated properly immediately
        # Try to set the radius only if top is not threaded
        if (
            self.params["r_constrained"]
            and self.params["type_top"] not in rea.THREADED_TOPS
        ):
            self.cad_obj.r = self.params["r"]
        else:
            self.cad_obj.r_constrained = False

        # If temporary reactor was built, update cad_object
        self.l.debug("Reactor, buildModule w cad_obj")

        self.cad_obj.volume = self.params["volume"]
        self.cad_obj.type_top = self.params["type_top"]
        self.cad_obj.type_bottom = self.params["type_bottom"]
        self.cad_obj.d_can = self.params["d_can"]
        self.cad_obj.align_top_strategy = self.params["align_top_strategy"]

        # Try to build I/Os
        try:
            self.buildInputs(self.cad_obj)
            self.buildOutputs(self.cad_obj)
            self.buildTopInlets(self.cad_obj)
        except Exception as e:
            mes = f"Impossible to create I/Os: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "I/Os error", mes, QtWidgets.QMessageBox.Ok
            )
            return False

        # Try to refresh assembly
        try:
            self.parent.refresh()
        except Exception as e:
            mes = f"Impossible to refresh reactor: {e}"
            self.l.error(mes, exc_info=True)
            QtWidgets.QMessageBox.critical(
                self, "Refresh error", mes, QtWidgets.QMessageBox.Ok
            )

        return True

    def _buildCADObject(self) -> rea:

        """
        Use the parameters read from the form to build the CAD object. Does
        nothing else, no GUI task here
        """

        volume = self.params["volume"]
        type_top = self.params["type_top"]
        type_bottom = self.params["type_bottom"]

        # Advanced settings
        d_can = self.params["d_can"]
        r = self.params["r"]
        r_constrained = self.params["r_constrained"]
        align_top_strategy = self.params["align_top_strategy"]

        reactor = rea(
            volume,
            type_top,
            type_bottom,
            d_can=d_can,
            align_top_strategy=align_top_strategy,
        )

        # Try to set the radius only if top is not threaded
        if r_constrained and type_top not in rea.THREADED_TOPS:
            reactor.r = r

        return reactor

    def initUI(self):

        """Handles the display"""

        if self.cad_obj is None:
            self.setWindowTitle("Add a reactor module")
        else:
            self.setWindowTitle("Modify reactor module")

        # Build the basic settings tab
        widget_form_basic = self._buildBasicSettings()

        # Build the advanced settings tab
        widget_form_advanced = self._buildAdvancedSettings()

        # Build the IO tab
        widget_io = self._buildIOTab()

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)

        self.tabs = QtWidgets.QTabWidget(self)

        self.tabs.addTab(widget_form_basic, "Basic")
        self.tabs.addTab(widget_form_advanced, "Advanced")
        self.tabs.addTab(widget_io, "Inputs/Outputs")

        self.ok_button = QtWidgets.QPushButton("OK", self)
        self.del_button = QtWidgets.QPushButton("Delete", self)
        split_butttons = QtWidgets.QHBoxLayout()
        split_butttons.addWidget(self.ok_button)
        split_butttons.addWidget(self.del_button)

        # if new object only display ok button
        if self.cad_obj is None:
            self.del_button.setVisible(False)

        # self.vbox_global.addLayout(fbox)
        self.vbox_global.addWidget(self.tabs)
        self.vbox_global.addLayout(split_butttons)

        self.setLayout(self.vbox_global)
        self.show()

    def _buildBasicSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the basic settings"""

        widget_form_basic = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_basic.setLayout(fbox)

        # Spin box to specify reactor's volume
        self.spin_vol = QtWidgets.QDoubleSpinBox(self)

        # Set default and bound values
        self.spin_vol.setValue(20)
        self.spin_vol.setMinimum(0)
        self.spin_vol.setMaximum(200)

        # Combo box to choose top of reactor
        self.combo_top = QtWidgets.QComboBox(self)
        self.combo_top.addItems(rea.TOPS)

        # Combo box to choose bottom of reactor
        # Set default to round internal
        self.combo_bottom = QtWidgets.QComboBox(self)
        list_bottom = rea.BOTTOMS
        self.combo_bottom.addItems(list_bottom)
        self.combo_bottom.setCurrentIndex(2)

        # Add widgets to form layout
        fbox.addRow("Reaction volume (mL): ", self.spin_vol)
        fbox.addRow("Top type: ", self.combo_top)
        fbox.addRow("Bottom type: ", self.combo_bottom)

        return widget_form_basic

    def _buildAdvancedSettings(self) -> QtWidgets.QWidget:

        """Build the widget for the advanced settings"""

        widget_form_advanced = QtWidgets.QWidget(self)
        fbox = QtWidgets.QFormLayout()
        widget_form_advanced.setLayout(fbox)

        # Spin box for internal radius of reactor
        # Disabled by default
        self.spin_r = QtWidgets.QDoubleSpinBox(self)
        self.spin_r.setEnabled(False)

        # Check box to change the state of radius constrained
        self.check_r_constrained = QtWidgets.QCheckBox(self)

        # Internal pipe diameter
        # Set default
        self.spin_d_can = QtWidgets.QDoubleSpinBox(self)
        self.spin_d_can.setValue(cst.D_CAN)

        # Combo box to choose how to align the top of the reactor
        self.combo_align_top_stratgey = QtWidgets.QComboBox(self)
        self.combo_align_top_stratgey.addItems(rea.ALIGN_TOP_STRATEGIES)

        # Add widgets to form layout
        fbox.addRow("Reactor radius: ", self.spin_r)
        fbox.addRow("Radius constrained: ", self.check_r_constrained)
        fbox.addRow("Diameter internal pipe: ", self.spin_d_can)
        fbox.addRow("Align top strategy: ", self.combo_align_top_stratgey)

        return widget_form_advanced

    def _buildIOTab(self) -> QtWidgets.QWidget:

        """Build the IO tab"""

        # I/O widget with layout
        widget_io = QtWidgets.QWidget(self)
        vbox_io = QtWidgets.QVBoxLayout()
        widget_io.setLayout(vbox_io)

        self.table_io = QtWidgets.QTableWidget(0, 3, self)
        self.table_io.setHorizontalHeaderLabels(["Name", "Type", "Connected"])
        self.table_io.setEditTriggers(QtWidgets.QAbstractItemView.NoEditTriggers)

        # Stretch header to fill entire space
        self.table_io.horizontalHeader().setSectionResizeMode(
            QtWidgets.QHeaderView.Stretch
        )

        # Disable header resizing
        self.table_io.verticalHeader().setSectionResizeMode(QtWidgets.QHeaderView.Fixed)

        # Selection selects entire rows
        self.table_io.setSelectionBehavior(QtWidgets.QAbstractItemView.SelectRows)

        # Only select one row at a time
        self.table_io.setSelectionMode(QtWidgets.QAbstractItemView.SingleSelection)

        # Button to add io (opens a dialog box)
        self.button_add_io = QtWidgets.QPushButton("Add I/O", self)

        # Button to remove selected io
        self.button_delete_io = QtWidgets.QPushButton("Delete I/O", self)

        # Add the widgets to the layout
        vbox_io.addWidget(self.table_io)
        vbox_io.addWidget(self.button_add_io)
        vbox_io.addWidget(self.button_delete_io)

        return widget_io


if __name__ == "__main__":
    print(os.path.dirname(os.path.dirname(os.path.dirname(os.path.abspath(__file__)))))
    app = QtWidgets.QApplication(sys.argv)
    obj = ChemModule()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8


import sys
import os
from typing import Tuple
from pathlib import Path

import chemscad.app_constants as cst


def getRightDirs() -> Tuple[str, str]:

    """Get the DATA_PATH and the resource_dir pathes.
    DATA_PATH is on the user side if CB is frozen"""

    root_path = os.path.dirname(os.path.abspath(__file__))

    if getattr(sys, "frozen", False):
        # resource_dir = os.path.dirname(os.path.realpath(sys.argv[0]))
        resource_dir = sys._MEIPASS
        DATA_PATH = cst.DATA_PATH
    else:
        resource_dir = root_path
        DATA_PATH = root_path

    return resource_dir, DATA_PATH


def getVersion() -> str:

    """This is a function JP copy/pasted from ChemBrows
    Therefore it needs to be rewritten, but it is never
    used in the main code."""

    resource_dir, DATA_PATH = getRightDirs()

    with open(
        os.path.join(resource_dir, "config/version.txt"), "r", encoding="utf-8"
    ) as version_file:

        version = version_file.read().strip()

    return version


def cleanRendersDir(logger):

    """
    Clean the renders directory. Will be called when ChemCAD is about to
    be closed
    https://stackoverflow.com/questions/185936/how-to-delete-the-contents-of-a-folder-in-python
    """

    logger.debug("Cleaning rendering directory")

    resource_dir, DATA_PATH = getRightDirs()

    folder = os.path.join(DATA_PATH, "renders")

    for file_del in os.listdir(folder):
        file_path = os.path.join(folder, file_del)
        try:
            if os.path.isfile(file_path):
                os.unlink(file_path)
                logger.debug(f"Removing {file_del} from render dir")
        except Exception as e:
            logger.error(e, exc_info=True)


if __name__ == "__main__":
    pass



#!/usr/bin/python
# coding: utf-8

import os
import sys
from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Union, List, Dict, Optional
from ccad import ReactorCAD
from ccad.object import ObjectCAD

from chemscad.log import MyLog


class AlignTops(QtWidgets.QDialog):

    """
    Create a modal window to select couples of tops of modules to align. The
    user will choose through a combo box. Will send the couples to the
    parent's assembly
    """

    def __init__(self, parent=None):

        super(AlignTops, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        if parent is None:
            self.l = MyLog("activity.log")
            self.test = True
        else:
            self.l = self.parent.l
            self.options = self.parent.options
            self.test = False

        # List of tuples of combo boxes
        self.list_boxes_align: List[QtWidgets.QComboBox] = list()

        self.initUI()
        self.defineSlots()

        self.loadValues()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # Button to add line
        self.button_add_line.clicked.connect(self.addLine)

        # User finished selecting tops to align
        self.ok_button.clicked.connect(self.readForm)

    def loadValues(self):

        """
        Load the values/name that will be used in combo boxes
        """

        if self.test:
            return

        self.dict_objects: Dict[str, ObjectCAD] = self.parent.ass.reactors_names

    def restoreParams(self):

        """
        Will restore the tops that need to be aligned, from Assembly.
        Will create a ligne of combo boxes for each couple
        """

        if self.test:
            return

        # For all the couples of tops in the assembly
        for couple in self.parent.ass.list_align_tops:
            mod_mov, mod_target = couple

            # Find the name of the objects
            for name, obj in self.dict_objects.items():
                if mod_mov is obj:
                    name_mod_move = name
                elif mod_target is obj:
                    name_mod_target = name

            # Add a line of combo boxes with the names
            self.addLine(False, name_mod_move, name_mod_target)

    def readForm(self):

        """
        Slot called when OK button is pressed. Will read the lines of the
        form, and send the tops to align to the Assembly of the parent object
        """

        # Clean the list of tops to align in assembly, we'll rebuild it below
        self.parent.ass.cleanAlignTops()

        # For all the combo boxes in every line
        for couple in self.list_boxes_align:
            combo_mov, combo_target = couple

            # Get the name of the objects
            name_mod_mov = combo_mov.currentText()
            name_mod_target = combo_target.currentText()

            # Get the object from their name
            mod_mov = self.dict_objects[name_mod_mov]
            mod_target = self.dict_objects[name_mod_target]

            # Check aligning mod_mov and mod_target is possible
            possible = self._checkAlignement(
                name_mod_mov, mod_mov, name_mod_target, mod_target
            )

            # If aligning tops is not possible, pass this couple
            if not possible:
                continue

            # Add the couple to the list of tops to align in assembly
            self.parent.ass.addAlignTops(mod_mov, mod_target)

        # General refresh from main window
        self.parent.refresh()

        self.close()

    def _checkAlignement(
        self,
        name_mod_mov: str,
        mod_mov: ReactorCAD,
        name_mod_target: str,
        mod_target: ReactorCAD,
    ) -> bool:

        """Check if the module can be aligned"""

        # TODO: check mod_mov is not present twice as moving module

        if mod_mov is mod_target:
            mes = f"You can't align module {mod_target} on itself. Ligne ignored"
            QtWidgets.QMessageBox.critical(
                self, "Selection error", mes, QtWidgets.QMessageBox.Ok
            )

            return False

        return True

    def addLine(
        self, 
        clicked: bool,
        name_mod_move: Optional[str] = None, 
        name_mod_target: Optional[str] = None
    ):

        """
        Slot called when the user presses the addLine button. 
        Argument clicked is set by PyQT when the button is clicked.
        Add a line for a new couple of tops to align
        """

        print(name_mod_move)

        # Create 2 new combo boxes
        combo_mod_mov = QtWidgets.QComboBox()
        combo_mod_target = QtWidgets.QComboBox()

        if not self.test:
            combo_mod_mov.addItems(self.dict_objects.keys())
            combo_mod_target.addItems(self.dict_objects.keys())

        # If addLine is called overloaded, it's called from restoreParams,
        # restore the tops to be aligned
        if name_mod_move is not None:
            combo_mod_mov.setCurrentText(name_mod_move)
        if name_mod_target is not None:
            combo_mod_target.setCurrentText(name_mod_target)

        # Append the 2 new boxes to the list of combo boxes
        self.list_boxes_align.append((combo_mod_mov, combo_mod_target))

        # Create 2 labels to identify the combo boxes
        label_mov = QtWidgets.QLabel("Align top of ")
        label_target = QtWidgets.QLabel("On top of ")

        button_delete = QtWidgets.QPushButton("Delete", self)

        hbox_line = QtWidgets.QHBoxLayout()
        hbox_line.addWidget(label_mov)
        hbox_line.addWidget(combo_mod_mov)
        hbox_line.addWidget(label_target)
        hbox_line.addWidget(combo_mod_target)
        hbox_line.addWidget(button_delete)

        button_delete.clicked.connect(
            lambda: self.deleteLine(hbox_line, combo_mod_mov, combo_mod_target)
        )

        self.vbox_lines_couples.addLayout(hbox_line)

    def deleteLine(
        self,
        hbox: QtWidgets.QHBoxLayout,
        combo_mod_mov: QtWidgets.QComboBox,
        combo_mod_target: QtWidgets.QComboBox,
    ):

        """
        Delete a line of combo boxes when user presses delete button
        of the line
        """

        # Switch boolean for moving module
        mod_mov = self.dict_objects[combo_mod_mov.currentText()]
        mod_target = self.dict_objects[combo_mod_target.currentText()]

        # Remove couple from assembly. Let assembly do other operations
        self.parent.ass.removeCoupleAlignTops((mod_mov, mod_target))

        self.list_boxes_align.remove((combo_mod_mov, combo_mod_target))
        self.clearLayout(hbox)

    def clearLayout(self, layout):

        """Method to erase the widgets from a layout"""

        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                widget = item.widget()
                if widget is not None:
                    widget.deleteLater()

                    QtWidgets.qApp.processEvents()
                else:
                    self.clearLayout(item.layout())

    def initUI(self):

        """Handles the display"""

        self.setWindowTitle("Tops to align")
        # self.setMinimumSize(200, 50)

        self.vbox_lines_couples = QtWidgets.QVBoxLayout()

        self.button_add_line = QtWidgets.QPushButton("Add line", self)
        self.ok_button = QtWidgets.QPushButton("OK", self)

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)
        self.vbox_global.addLayout(self.vbox_lines_couples)
        self.vbox_global.addWidget(self.button_add_line)
        self.vbox_global.addWidget(self.ok_button)

        self.setLayout(self.vbox_global)
        self.show()


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    obj = AlignTops()
    sys.exit(app.exec_())



#!/usr/bin/python
# coding: utf-8

import os
import sys
from PyQt5 import QtGui, QtCore, QtWidgets
from typing import Union, List, Dict, Optional

from chemscad.log import MyLog

from ccad.object import ObjectCAD


class AlignFilters(QtWidgets.QDialog):

    """
    Create a modal window to select couples of modules with filters to
    align. The user will choose through a combo box. Will send the couples
    to the parent's assembly
    """

    def __init__(self, parent=None) -> None:

        super(AlignFilters, self).__init__(parent)

        self.setModal(True)

        self.setAttribute(QtCore.Qt.WA_DeleteOnClose)

        self.parent = parent

        if parent is None:
            self.l = MyLog("activity.log")
            self.test = True
        else:
            self.l = self.parent.l
            self.options = self.parent.options
            self.test = False

        # List of tuples of combo boxes
        self.list_boxes_align: List[QtWidgets.QComboBox] = list()

        self.initUI()
        self.defineSlots()

        self.loadValues()
        self.restoreParams()

    def defineSlots(self):

        """Establish the slots"""

        # Button to add line
        self.button_add_line.clicked.connect(self.addLine)

        # User finished selecting filters to align
        self.ok_button.clicked.connect(self.readForm)

    def loadValues(self):

        """
        Load the values/name that will be used in combo boxes
        """

        if self.test:
            return

        # Use a dict comprehension to only get objects with filter
        self.dict_objects: Dict[str, ObjectCAD] = {
            name: obj
            for name, obj in self.parent.ass.reactors_names.items()
            if hasattr(obj, "z_top_filter")
        }

    def restoreParams(self):

        """
        Will restore the filters that need to be aligned, from Assembly.
        Will create a ligne of combo boxes for each couple
        """

        if self.test:
            return

        # For all the couples of filters in the assembly
        for couple in self.parent.ass.list_align_filters:
            mod_mov, mod_target = couple

            # Find the name of the objects
            for name, obj in self.dict_objects.items():
                if mod_mov is obj:
                    name_mod_move = name
                elif mod_target is obj:
                    name_mod_target = name

            # Add a line of combo boxes with the names
            self.addLine(name_mod_move, name_mod_target)

    def readForm(self):

        """
        Slot called when OK button is pressed. Will read the lines of
        the form, and send the filters to align to the Assembly of the
        parent object
        """

        # Clean the list of filters to align in assembly, we'll rebuild
        # it below
        self.parent.ass.cleanAlignFilters()

        # For all the combo boxes in every line
        for i, couple in enumerate(self.list_boxes_align):
            combo_mov, combo_target = couple

            # Get the object from their name
            mod_mov = self.dict_objects[combo_mov.currentText()]
            mod_target = self.dict_objects[combo_target.currentText()]

            if mod_mov is mod_target:
                mes = f"Module selected twice on line {i + 1}. Line ignored"
                QtWidgets.QMessageBox.critical(
                    self, "Selection error", mes, QtWidgets.QMessageBox.Ok
                )

                continue

            # Try to align by by-passing assembly. If it doesn't fail,
            # tell the assembly to align
            try:
                mod_mov.alignFilterToHeight(mod_target.z_top_filter)
            except Exception as e:
                self.l.error("Problem with align_filters: {e}", exc_info=True)

                name_mod_mov = combo_mov.currentText()
                name_mod_target = combo_target.currentText()

                mes = "Impossible to align filters from {} and {}: {}"
                mes = mes.format(name_mod_mov, name_mod_target, e)
                QtWidgets.QMessageBox.critical(
                    self, "Refresh error", mes, QtWidgets.QMessageBox.Ok
                )
                continue

            # Add the couple to the list of filters to align in assembly
            self.parent.ass.addAlignFilters(mod_mov, mod_target)

        self.parent.refresh()
        self.close()

    def addLine(
        self, name_mod_move: Optional[str] = None, name_mod_target: Optional[str] = None
    ):

        """
        Slot called when the user presses the addLine button. Add a line
        for a new couple of filters to align
        """

        # Create 2 new combo boxes
        combo_mod_mov = QtWidgets.QComboBox()
        combo_mod_target = QtWidgets.QComboBox()

        if not self.test:
            combo_mod_mov.addItems(self.dict_objects.keys())
            combo_mod_target.addItems(self.dict_objects.keys())

        # If addLine is called overloaded, it's called from restoreParams,
        # restore the filters to be aligned
        if name_mod_move is not None:
            combo_mod_mov.setCurrentText(name_mod_move)
        if name_mod_target is not None:
            combo_mod_target.setCurrentText(name_mod_target)

        # Append the 2 new boxes to the list of combo boxes
        self.list_boxes_align.append((combo_mod_mov, combo_mod_target))

        # Create 2 labels to identify the combo boxes
        label_mov = QtWidgets.QLabel("Align filter of ")
        label_target = QtWidgets.QLabel("On filter of ")

        button_delete = QtWidgets.QPushButton("Delete", self)

        hbox_line = QtWidgets.QHBoxLayout()
        hbox_line.addWidget(label_mov)
        hbox_line.addWidget(combo_mod_mov)
        hbox_line.addWidget(label_target)
        hbox_line.addWidget(combo_mod_target)
        hbox_line.addWidget(button_delete)

        button_delete.clicked.connect(
            lambda: self.deleteLine(hbox_line, combo_mod_mov, combo_mod_target)
        )

        self.vbox_lines_couples.addLayout(hbox_line)

    def deleteLine(
        self,
        hbox: QtWidgets.QHBoxLayout,
        combo_mod_mov: QtWidgets.QComboBox,
        combo_mod_target: QtWidgets.QComboBox,
    ):

        """
        Delete a line of combo boxes when user presses delete button
        of the line
        """

        # Switch boolean for moving module
        mod_mov = self.dict_objects[combo_mod_mov.currentText()]
        mod_target = self.dict_objects[combo_mod_target.currentText()]

        # Remove couple from assembly. Let assembly do other operations
        self.parent.ass.removeCoupleAlignFilters((mod_mov, mod_target))

        self.list_boxes_align.remove((combo_mod_mov, combo_mod_target))
        self.clearLayout(hbox)

    def clearLayout(self, layout):

        """Method to erase the widgets from a layout"""

        if layout is not None:
            while layout.count():
                item = layout.takeAt(0)
                widget = item.widget()
                if widget is not None:
                    widget.deleteLater()

                    QtWidgets.qApp.processEvents()
                else:
                    self.clearLayout(item.layout())

    def initUI(self):

        """Handles the display"""

        self.setWindowTitle("Filters to align")
        # self.setMinimumSize(200, 50)

        self.vbox_lines_couples = QtWidgets.QVBoxLayout()

        self.button_add_line = QtWidgets.QPushButton("Add line", self)
        self.ok_button = QtWidgets.QPushButton("OK", self)

        # ----------------- ASSEMBLING ----------------------------------------

        self.vbox_global = QtWidgets.QVBoxLayout(self)
        self.vbox_global.addLayout(self.vbox_lines_couples)
        self.vbox_global.addWidget(self.button_add_line)
        self.vbox_global.addWidget(self.ok_button)

        self.setLayout(self.vbox_global)
        self.show()


if __name__ == "__main__":

    app = QtWidgets.QApplication(sys.argv)
    obj = AlignFilters()
    sys.exit(app.exec_())
